(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: translmod.ml,v 1.28 1999/02/25 14:02:44 xleroy Exp $ *)

(* Translation from typed abstract syntax to lambda terms,
   for the module language *)

open Misc
open Asttypes
open Path
open Types
open Typedtree
open Lambda
open Translobj
open Translcore
open Translclass

let get_name id = Lprim(Pfield 0, [Lvar id])

let name_id id = (* create name *)
  Lconst(Const_block (0, [Const_base (Const_string (Ident.name id));
			  Const_base (Const_int (Ident.stamp id))])) 

let name_dot parname myid = (* create name.param *)
  Lprim(Pmakeblock(1, Immutable), [ parname; 
			Lconst (Const_base (Const_string (Ident.name myid)));
			Lconst (Const_base (Const_int (Ident.stamp myid))) ])

let name_app name mdl = (* create name(param) *)
  Lprim(Pmakeblock(2, Immutable), [ name; get_name mdl ])

let name_dummy = Lconst(Const_block (0, [Const_base (Const_string "dummy");
					 Const_base (Const_int 0)]))

(* Compile a coercion *)

let rec apply_coercion restr arg =
  match restr with
    Tcoerce_none ->
      arg
  | Tcoerce_structure pos_cc_list ->
(* prerr_endline "*** Tcoerce_structure"; *)
      name_lambda arg (fun id ->
        Lprim(Pmakeblock(0, Immutable),
	      Lprim(Pfield 0, [Lvar id]) :: (* JPF Modname *)
              List.map (apply_coercion_field id) pos_cc_list))
  | Tcoerce_functor(cc_arg, cc_res) ->
(* prerr_endline "*** Tcoerce_functor"; *)
      let param = Ident.create "funarg" in
      name_lambda arg (fun id ->
	  Lprim(Pmakeblock(0, Immutable), [
	      Lprim(Pfield 0, [Lvar id]); (* JPF Modname *)
	      Lfunction(Curried, [param],
			apply_coercion cc_res
			  (Lapply( Lprim(Pfield 1,[Lvar id]), (* JPF Functor body *) 
				  [apply_coercion cc_arg (Lvar param)]))) ]))
  | Tcoerce_primitive p ->
      transl_primitive p

and apply_coercion_field id (pos, cc) =
  apply_coercion cc (Lprim(Pfield (pos + 1) (* JPF *), [Lvar id]))
(*  apply_coercion cc (Lprim(Pfield pos, [Lvar id])) *)

(* Compose two coercions
   apply_coercion c1 (apply_coercion c2 e) behaves like
   apply_coercion (compose_coercions c1 c2) e. *)

let rec compose_coercions c1 c2 =
  match (c1, c2) with
    (Tcoerce_none, c2) -> c2
  | (c1, Tcoerce_none) -> c1
  | (Tcoerce_structure pc1, Tcoerce_structure pc2) ->
      let v2 = Array.of_list pc2 in
      Tcoerce_structure
        (List.map
          (function (p1, Tcoerce_primitive p) ->
                      (p1, Tcoerce_primitive p)
                  | (p1, c1) ->
                      let (p2, c2) = v2.(p1) in (p2, compose_coercions c1 c2))
             pc1)
  | (Tcoerce_functor(arg1, res1), Tcoerce_functor(arg2, res2)) ->
      Tcoerce_functor(compose_coercions arg2 arg1,
                      compose_coercions res1 res2)
  | (_, _) ->
      fatal_error "Translmod.compose_coercions"

(* Record the primitive declarations occuring in the module compiled *)

let primitive_declarations = ref ([] : string list)

(* Keep track of the root path (from the root of the namespace to the
   currently compiled module expression).  Useful for naming exceptions. *)

let global_path glob = Some(Pident glob)
let functor_path path param =
  match path with
    None -> None
  | Some p -> Some(Papply(p, Pident param))
let field_path path field =
  match path with
    None -> None
  | Some p -> Some(Pdot(p, Ident.name field, Path.nopos))
let root_path_name = function
    Some p -> Path.string_of p
  | None -> "none"

(* Generic extensions *)
(***************************************************** UNDER CONSTRUCTION 
let default_holder_ident id = Ident.create (Ident.name id ^ "%def")
let default_idents bound_idents =
  List.map (fun id -> default_holder_ident id) bound_idents

let create_defaults default_holder_idents default_error_idents rem =
  let rec create_default_errors rem = 
    let sub id rem =
      Llet (Strict, 
	    id, 
  	    Lprim(Pmakeblock (0, Mutable), [
  	     Lfunction (Curried, [Ident.create "dummy"], Lprim(Praise, 
  	      [Lprim(Pmakeblock(0, Immutable),
  		[ transl_path Predef.path_match_failure;
  		  Lconst(Const_block(0, [
  		    Const_base(Const_string !Location.input_name);
  		    Const_base(Const_int 0 (* loc.loc_start *) );
  		    Const_base(Const_int 1 (* loc.loc_end *) )]))])]))]),
  		  rem)
    in
    function 
	[id] -> sub id rem
      | id :: ids -> sub id (create_default_errors rem ids)
  in
  let rec create_default_holders rem = fun dhs des ->
    let sub id id' rem =
      Llet(Strict, id, Lprim(Pmakeblock (0, Mutable), [Lvar id']), rem)
    in
    match dhs, des with
      [id], [id'] -> sub id id' rem
    | id :: ids, id' :: ids' -> 
	sub id id' (create_default_holders rem ids ids')
  in
  create_default_errors 
    ( create_default_holders rem default_holder_idents default_error_idents )
    default_error_idents
**************************************************************************)

(* Compile a module expression *)

(*
let modnameid = Ident.create "*mname*"
*)

let rec transl_module cc rootpath name mexp =
  match mexp.mod_desc with
    Tmod_ident path ->
      apply_coercion cc (transl_path path)
  | Tmod_structure str ->
      transl_structure [] cc rootpath name str
  | Tmod_functor(param, mty, body) ->
      let bodypath = functor_path rootpath param in
      begin match cc with
        Tcoerce_none ->
	  Lprim(Pmakeblock (0, Immutable), [
	    name;
	    Lfunction(Curried, [param], 
		      transl_module Tcoerce_none bodypath 
			(name_app name param) body)])
      | Tcoerce_functor(ccarg, ccres) ->
          let param' = Ident.create "funarg" in
          Lprim(Pmakeblock (0, Immutable), [
	    name;
	    Lfunction(Curried, [param'],
		  Llet(Alias, param, apply_coercion ccarg (Lvar param'),
			       transl_module ccres bodypath 
			         (name_app name param') body))])
      | _ ->
          fatal_error "Translmod.transl_module"
      end
  | Tmod_apply(funct, arg, ccarg) ->
      apply_coercion cc
        (Lapply(Lprim(Pfield 1, 
		      [transl_module Tcoerce_none None name_dummy funct]),
                [transl_module ccarg None name_dummy arg]))
  | Tmod_constraint(arg, mty, ccarg) ->
      transl_module (compose_coercions cc ccarg) rootpath name arg

and transl_structure fields cc rootpath name str = 
  let modnameid = Ident.create "*mname*" in
  Llet (Strict, modnameid, name, 
	transl_structure_core fields cc rootpath modnameid str)

and transl_structure_core fields cc rootpath modnameid = function
    [] ->
(* JPF: Hack for module name *)      

      begin match cc with
        Tcoerce_none ->
          Lprim(Pmakeblock(0, Immutable),
		(Lvar modnameid) ::
		List.map (fun id -> Lvar id) (List.rev fields))
      | Tcoerce_structure pos_cc_list ->
          let v = Array.of_list (List.rev fields) in
          Lprim(Pmakeblock(0, Immutable),
		(Lvar modnameid) ::
                List.map
                  (fun (pos, cc) ->
                    match cc with
                      Tcoerce_primitive p -> transl_primitive p
                    | _ -> apply_coercion cc (Lvar v.(pos)))
                  pos_cc_list)
      | _ ->
          fatal_error "Translmod.transl_structure_core"
      end
(*
      begin match cc with
        Tcoerce_none ->
          Lprim(Pmakeblock(0, Immutable),
		 List.map (fun id -> Lvar id) (List.rev fields))
      | Tcoerce_structure pos_cc_list ->
          let v = Array.of_list (List.rev fields) in
          Lprim(Pmakeblock(0, Immutable),
                List.map
                  (fun (pos, cc) ->
                    match cc with
                      Tcoerce_primitive p -> transl_primitive p
                    | _ -> apply_coercion cc (Lvar v.(pos)))
                  pos_cc_list)
      | _ ->
          fatal_error "Translmod.transl_structure_core"
      end
*)
  | Tstr_eval (expr, scm, typecase, flowptr) :: rem ->
      Lsequence(transl_eval flowptr scm expr, transl_structure_core fields cc rootpath modnameid rem)
  | Tstr_value(rec_flag, pat_expr_list) :: rem ->
      let ext_fields = rev_let_bound_idents pat_expr_list @ fields in
      transl_let rec_flag pat_expr_list
                 (transl_structure_core ext_fields cc rootpath modnameid rem)
  | Tstr_generic(Nonextensible, rec_flag, env', defs) :: rem ->
      let ext_fields = generic_bound_idents defs @ fields in
      transl_generic rec_flag env' defs
                 (transl_structure_core ext_fields cc rootpath modnameid rem)
(**************************************************** UNDER CONSTRUCTION
  | Tstr_generic(Extensible, rec_flag, defs) :: rem ->
      prerr_endline "Extensible";

      let gbounds = generic_bound_idents defs in
      let errors = List.map (fun _ -> Ident.create "error") gbounds in
      let ext_fields = holders @ gbounds @ fields in
      
      create_defaults holders errors
        (transl_generic rec_flag (List.map2 (fun def def_error -> def, 
	  Some (Lprim( Pfield 0, [Lvar def_error]))) defs errors)
	   (transl_structure_core ext_fields cc rootpath rem)) 
  | Tstr_genextend(path, dpath, typ, Expand, cases) :: rem ->
      prerr_endline "Extend!!!";
      let error_ident = Ident.create "error" in
      let ext_fields = error_ident :: fields in
      (* let new_default = ref (!!f%def) *)
      Llet (Strict, error_ident,
	    Lprim(Pmakeblock (0, Mutable), [
              Lprim(Pfield 0, [ Lprim(Pfield 0, [transl_path dpath] )])]),
      (* let x = generic _ = case t of cases | _ -> !new_default *)
(* BUG defined in translcore also *)
let dummy_loc =   {Location.loc_start= 0; Location.loc_end= 1 } (* dummy *)
in
      let xid = Ident.create "x" in
      let xpat = { pat_desc= Tpat_var xid;
		   pat_loc= dummy_loc;
		   pat_type= Ctype.none; (* dummy *)
		   pat_env= Env.empty } (* ??? *)
      in
      transl_generic Nonrecursive [ 
        (xpat, (typ, cases)), Some (Lprim( Pfield 0, [Lvar error_ident])) ] (
      Lsequence (
        (* !f%def := x *)
        Lprim(Psetfield (0, true), [ Lprim(Pfield 0, [transl_path dpath]);
				     Lvar xid ]),
        (* f%def := new_default *)
        Lsequence (
          Lprim(Psetfield (0, true), [ transl_path dpath; Lvar error_ident ]),
          transl_structure_core ext_fields cc rootpath modnameid rem))))
	      
*)
  | Tstr_primitive(id, descr) :: rem ->
      begin match descr.val_kind with
        Val_prim p -> primitive_declarations :=
                        p.Primitive.prim_name :: !primitive_declarations
      | _ -> ()
      end;
      transl_structure_core fields cc rootpath modnameid rem
  | Tstr_type(decls) :: rem ->
      (* JPF *)
(*
      (* sort... *)
      let sorted = ref [] in
      let rec sorter (id, decl) =
	try List.assoc id !sorted; () 
	with Not_found -> 
	  match decl.type_digest with
	  | DSource s -> 
	      List.iter (function
		  Pident i ->
		    begin try 
		      let decl' = List.assoc i decls in
		      sorter (i, decl')
		    with Not_found -> 
                        (* previously defined *)
		      ()
		    end
		| _ -> ()) s.regular;
	      sorted := (id, decl) :: !sorted
	  | _ -> sorted := (id, decl) :: !sorted
      in
      List.iter sorter decls;
      let decls = List.rev !sorted in
*)
      let idents = List.map fst decls in
      let defs = List.map (Bdigest.transl_type_info_creation 
			     (Lvar modnameid)) decls in
      List.fold_right (fun (id, def) st -> Llet(Strict, id, def, st)) 
	defs (transl_structure_core (idents @ fields) cc rootpath modnameid rem)
  | Tstr_exception(id, decl) :: rem ->
      Llet(Strict, id, transl_exception id (field_path rootpath id) decl,
           transl_structure_core (id :: fields) cc rootpath modnameid rem)
  | Tstr_module(id, modl) :: rem ->
      Llet(Strict, id,
           transl_module Tcoerce_none (field_path rootpath id) 
	     (name_dot (Lvar modnameid) id) modl,
           transl_structure_core (id :: fields) cc rootpath modnameid rem)
  | Tstr_modtype(id, decl) :: rem ->
      transl_structure_core fields cc rootpath modnameid rem
  | Tstr_open path :: rem ->
      transl_structure_core fields cc rootpath modnameid rem
  | Tstr_class cl_list :: rem ->
      let ids = List.map (fun (i, _, _, _) -> i) cl_list in
      Lletrec(List.map
                (fun (id, arity, meths, cl) ->
                  (id, transl_class ids id arity meths cl))
                cl_list,
              transl_structure_core (List.rev ids @ fields) cc rootpath modnameid rem)
  | Tstr_cltype cl_list :: rem ->
      transl_structure_core fields cc rootpath modnameid rem

(* Update forward declaration in Translcore *)
let _ =
  Translcore.transl_module := transl_module

(* Compile an implementation *)

let transl_implementation module_name (str, cc) =
  reset_labels ();
  primitive_declarations := [];
  let module_id = Ident.create_persistent module_name in
  Lprim(Psetglobal module_id,
        [transl_label_init
            (transl_structure [] cc (global_path module_id) (name_id module_id) str)])

(* A variant of transl_structure used to compile toplevel structure definitions
   for the native-code compiler. Store the defined values in the fields
   of the global as soon as they are defined, in order to reduce register
   pressure.  Also rewrites the defining expressions so that they
   refer to earlier fields of the structure through the fields of
   the global, not by their names.
   "map" is a table from defined idents to (pos in global block, coercion).
   "prim" is a list of (pos in global block, primitive declaration). *)

let transl_store_structure glob map prims name str =
(* WARN ! this must be wrong ! *)
  let modid = Ident.create "*mname*" in

  let rec transl_store subst = function
    [] ->
      lambda_unit
  | Tstr_eval (expr, scm, typecase, flowptr) :: rem ->
      Lsequence(subst_lambda subst (transl_eval flowptr scm expr), 
		transl_store subst rem)
  | Tstr_value(rec_flag, pat_expr_list) :: rem ->
      let ids = let_bound_idents pat_expr_list in
      let lam = transl_let rec_flag pat_expr_list (store_idents ids) in
      Lsequence(subst_lambda subst lam,
                transl_store (add_idents ids subst) rem)
  | Tstr_generic(Nonextensible, rec_flag, env', defs) :: rem ->
      let ids = generic_bound_idents defs in
      let lam = transl_generic rec_flag env' defs
	  (store_idents ids)
      in
      Lsequence(subst_lambda subst lam,
		transl_store (add_idents ids subst) rem)
(****************************************************** UNDER CONSTRUCTION
  | Tstr_generic(Extensible, rec_flag, defs) :: rem ->
      prerr_endline "Extensible";
      let gbounds = generic_bound_idents defs in
      let errors = List.map (fun _ -> Ident.create "error") gbounds in 
      let all_idents = holders @ gbounds
      in
      create_defaults holders errors
        (transl_generic rec_flag (List.map2 (fun def def_error -> def, 
	  Some (Lprim( Pfield 0,[Lvar def_error]))) defs errors)
        (store_idents glob map all_idents
          (transl_store rem)))
***************************************************************************)
  | Tstr_primitive(id, descr) :: rem ->
      begin match descr.val_kind with
        Val_prim p -> primitive_declarations :=
                        p.Primitive.prim_name :: !primitive_declarations
      | _ -> ()
      end;
      transl_store subst rem
  | Tstr_type(decls) :: rem ->
      (* JPF *)
      let idents = List.map fst decls in
      let defs = List.map (Bdigest.transl_type_info_creation (Lvar modid))
	  decls in
      let lam = 
	List.fold_right (fun (id,def) st -> Llet(Strict, id, def, st))
	  defs (store_idents idents)
(*
        List.fold_right (fun (id, def) st -> Llet(Strict, id, def, st)) 
	  defs (store_idents glob map idents (transl_store subst rem))
*)
      in
      Lsequence(subst_lambda subst lam,
		transl_store (add_idents idents subst) rem)
(*
      transl_store subst rem
*)
  | Tstr_exception(id, decl) :: rem ->
      let lam = transl_exception id (field_path (global_path glob) id) decl in
      Lsequence(Llet(Strict, id, lam, store_ident id),
                transl_store (add_ident false id subst) rem)
  | Tstr_module(id, modl) :: rem ->
      let lam =
        transl_module Tcoerce_none (field_path (global_path glob) id) 
	  (name_dot (Lvar modid) id) modl in
      (* Careful: the module value stored in the global may be different
         from the local module value, in case a coercion is applied.
         If so, keep using the local module value (id) in the remainder of
         the compilation unit (add_ident true returns subst unchanged).
         If not, we can use the value from the global
         (add_ident true adds id -> Pgetglobal... to subst). *)
      Llet(Strict, id, subst_lambda subst lam,
        Lsequence(store_ident id, transl_store(add_ident true id subst) rem))
  | Tstr_modtype(id, decl) :: rem ->
      transl_store subst rem
  | Tstr_open path :: rem ->
      transl_store subst rem
  | Tstr_class cl_list :: rem ->
      let ids = List.map (fun (i, _, _, _) -> i) cl_list in
      let lam =
        Lletrec(List.map
                  (fun (id, arity, meths, cl) ->
                     (id, transl_class ids id arity meths cl))
                  cl_list,
                store_idents ids) in
      Lsequence(subst_lambda subst lam,
                transl_store (add_idents ids subst) rem)
  | Tstr_cltype cl_list :: rem ->
      transl_store subst rem

  and store_ident id =
    try
      let (pos, cc) = Ident.find_same id map in
      let init_val = apply_coercion cc (Lvar id) in
prerr_endline "store_ident?";
      Lprim(Psetfield(pos, false), [Lprim(Pgetglobal glob, []); init_val])
    with Not_found ->
      fatal_error("Translmod.store_ident: " ^ Ident.unique_name id)

  and store_idents idlist =
    make_sequence store_ident idlist

  and add_ident may_coerce id subst =
    try
      let (pos, cc) = Ident.find_same id map in
      match cc with
        Tcoerce_none ->
          Ident.add id (Lprim(Pfield pos, [Lprim(Pgetglobal glob, [])])) subst
      | _ ->
          if may_coerce then subst else assert false
    with Not_found ->
      assert false

  and add_idents idlist subst =
    List.fold_right (add_ident false) idlist subst

  and store_primitive (pos, prim) cont =
    Lsequence(Lprim(Psetfield(pos, false),
                    [Lprim(Pgetglobal glob, []); transl_primitive prim]),
              cont)

  in 
  List.fold_right store_primitive prims (transl_store Ident.empty str)

(* Build the list of value identifiers defined by a toplevel structure
   (excluding primitive declarations). *)

let rec defined_idents = function
    [] -> []
  | Tstr_eval (expr, scm, typecase, flowptr) :: rem -> defined_idents rem
  | Tstr_value(rec_flag, pat_expr_list) :: rem ->
      let_bound_idents pat_expr_list @ defined_idents rem
  | Tstr_generic(Nonextensible, rec_flag, env', defs) :: rem ->
      generic_bound_idents defs @ defined_idents rem
(*************************************** UNDER CONSTRUCTION
  | Tstr_generic(Extensible, rec_flag, defs) :: rem ->
      let gbounds = generic_bound_idents defs in
      let all_idents = holders @ gbounds in
      all_idents @ defined_idents rem
  | Tstr_genextend(path, dpath, typ, Expand, cases) :: rem ->
      let error_ident = Ident.create "error" in
      error_ident :: defined_idents rem
***************************)
  | Tstr_primitive(id, descr) :: rem -> defined_idents rem
  | Tstr_type decls :: rem -> 
      (* JPF *)
      map_end (fun (id,_) -> id) decls (defined_idents rem)
(*
      defined_idents rem
*)
  | Tstr_exception(id, decl) :: rem -> id :: defined_idents rem
  | Tstr_module(id, modl) :: rem -> id :: defined_idents rem
  | Tstr_modtype(id, decl) :: rem -> defined_idents rem
  | Tstr_open path :: rem -> defined_idents rem
  | Tstr_class cl_list :: rem ->
      List.map (fun (i, _, _, _) -> i) cl_list @ defined_idents rem
  | Tstr_cltype cl_list :: rem -> defined_idents rem

(* Transform a coercion and the list of value identifiers defined by
   a toplevel structure into a table [id -> (pos, coercion)],
   with [pos] being the position in the global block where the value of
   [id] must be stored, and [coercion] the coercion to be applied to it.
   A given identifier may appear several times
   in the coercion (if it occurs several times in the signature); remember
   to assign it the position of its last occurrence.
   Identifiers that are not exported are assigned positions at the
   end of the block (beyond the positions of all exported idents).
   Also compute the total size of the global block,
   and the list of all primitives exported as values. *)

let build_ident_map restr idlist =
  let rec natural_map pos map prims = function
    [] ->
      (map, prims, pos)
  | id :: rem ->
      natural_map (pos+1) (Ident.add id (pos, Tcoerce_none) map) prims rem in
  match restr with
    Tcoerce_none ->
prerr_endline "build_ident_map?";
(*      natural_map (* 0 *) 1 (* JPF *) Ident.empty [] idlist *)
      natural_map 0 Ident.empty [] idlist
  | Tcoerce_structure pos_cc_list ->
      let idarray = Array.of_list idlist in
      let rec export_map pos map prims undef = function
        [] ->
          natural_map pos map prims undef
      | (source_pos, Tcoerce_primitive p) :: rem ->
          export_map (pos + 1) map ((pos, p) :: prims) undef rem
      | (source_pos, cc) :: rem ->
          let id = idarray.(source_pos) in
          export_map (pos + 1) (Ident.add id (pos, cc) map)
                     prims (list_remove id undef) rem
(*      in export_map (* 0 *) 1  Ident.empty [] idlist pos_cc_list *)
      in
prerr_endline "build_ident_map?";
      export_map 0 Ident.empty [] idlist pos_cc_list
  | _ ->
      fatal_error "Translmod.build_ident_map"

(* Compile an implementation using transl_store_structure 
   (for the native-code compiler). *)

let transl_store_implementation module_name (str, restr) =
  reset_labels ();
  primitive_declarations := [];
  let module_id = Ident.create_persistent module_name in
  let (map, prims, size) = build_ident_map restr (defined_idents str) in
  (size, transl_label_init (transl_store_structure module_id map prims module_id str))

(* Compile a toplevel phrase *)


let toplevel_modname =
  (Lconst(Const_block (0, [Const_base (Const_string "*toplevel*");
			    Const_base (Const_int 0)])))

let transl_toplevel_item = function
    Tstr_eval (expr, scm, typecase, flowptr) ->
(*      transl_exp expr *)
      transl_eval flowptr scm expr
  | Tstr_value(rec_flag, pat_expr_list) ->
      let idents = let_bound_idents pat_expr_list in
      let lam =
        transl_let rec_flag pat_expr_list
          (make_sequence (fun id -> Lprim(Psetglobal id, [Lvar id])) idents) in
      List.iter Ident.make_global idents;
      lam
  | Tstr_generic(Nonextensible, rec_flag, env', defs) ->
      let idents = generic_bound_idents defs in
      let lam =
        transl_generic rec_flag env' defs
          (make_sequence (fun id -> Lprim(Psetglobal id, [Lvar id])) idents) in
      List.iter Ident.make_global idents;
      lam
(******************************************* UNDER CONSTRUCTION
  | Tstr_generic(Extensible, rec_flag, defs) ->
      prerr_endline "Extensible";
      let idents = generic_bound_idents defs in
      let errors = List.map (fun _ -> Ident.create "error") idents in
      let all_idents = holders @ idents in
      let lam =
	create_defaults holders errors
	 (transl_generic rec_flag (List.map2 (fun def def_error -> def, 
	   Some (Lprim (Pfield 0, [Lvar def_error]))) defs errors)
          (make_sequence (fun id -> Lprim(Psetglobal id, [Lvar id])) 
	     all_idents))
      in
      List.iter Ident.make_global all_idents;
      lam

  | Tstr_genextend(path, dpath, typ, Expand, cases) ->
      prerr_endline "Extend!!!";
      let error_ident = Ident.create "error" in
      let lam =
      (* let new_default = ref (!!f%def) *)
      Llet (Strict, error_ident,
	    Lprim(Pmakeblock (0, Mutable), [
              Lprim(Pfield 0, [ Lprim(Pfield 0, [transl_path dpath] )])]),
      (* let x = generic _ = case t of cases | _ -> !new_default *)
(* BUG defined in translcore also *)
let dummy_loc =   {Location.loc_start= 0; Location.loc_end= 1 } (* dummy *)
in
      let xid = Ident.create "x" in
      let xpat = { pat_desc= Tpat_var xid;
		   pat_loc= dummy_loc;
		   pat_type= Ctype.none; (* dummy *)
		   pat_env= Env.empty } (* ??? *)
      in
      transl_generic Nonrecursive [ 
        (xpat, (typ, cases)), Some (Lprim(Pfield 0, [Lprim (Pgetglobal error_ident, [])]))] (
        (* !f%def := x *)
      Lsequence (
        Lprim(Psetfield (0, true), [ Lprim(Pfield 0, [transl_path dpath]);
				     Lvar xid ]),
        (* f%def := new_default *)
        Lsequence (
          Lprim(Psetfield (0, true), [ transl_path dpath; Lvar error_ident ]),
          (make_sequence (fun id -> Lprim(Psetglobal id, [Lvar id])) [error_ident])))))
      in
      List.iter Ident.make_global [error_ident];
      lam
	      
********************************************************)
  | Tstr_primitive(id, descr) ->
      lambda_unit
  | Tstr_type(decls) ->
      (* JPF *)
      let idents = List.map fst decls in
      (* idents must be global before we create lambda *)
      List.iter (fun id -> Ident.make_global id) idents;
      let defs = List.map (Bdigest.transl_type_info_creation 
			     toplevel_modname) decls in
      let lam = make_sequence (fun (id,def) -> 
	Llet (Strict, id, def, Lprim(Psetglobal id, [Lvar id])))
	  defs
      in 
      lam
(*
      lambda_unit
*)
  | Tstr_exception(id, decl) ->
      Ident.make_global id;
      Lprim(Psetglobal id, [transl_exception id None decl])
  | Tstr_module(id, modl) ->
      Ident.make_global id;
      Lprim(Psetglobal id, [transl_module Tcoerce_none (global_path id) (name_dot toplevel_modname id) modl])
   (* Lprim(Psetglobal id, [transl_module Tcoerce_none None modl]) *)
  | Tstr_modtype(id, decl) ->
      lambda_unit
  | Tstr_open path ->
      lambda_unit
  | Tstr_class cl_list ->
      let ids = List.map (fun (i, _, _, _) -> i) cl_list in
      let lam =
        Lletrec(List.map
                  (fun (id, arity, meths, cl) ->
                     (id, transl_class ids id arity meths cl))
                  cl_list,
                make_sequence
                  (fun (id, _, _, _) -> Lprim(Psetglobal id, [Lvar id]))
                  cl_list)
      in
      List.iter (fun (id, _, _, _) -> Ident.make_global id) cl_list;
      lam
      
  | Tstr_cltype cl_list ->
      lambda_unit

let transl_toplevel_definition str =
  reset_labels ();
  transl_label_init (make_sequence transl_toplevel_item str)
