(* Digest *)

exception No_digest_info
exception Generic_rdigest_failure of string

external unsafe_string: string -> int -> int -> string = "md5_string"
let digest_string str =
  unsafe_string str 0 (String.length str)

let list_mem_nth k =
  let rec sub pos lst = 
    match lst with
    | x :: _ when x = k -> pos
    | _ :: xs -> sub (pos + 1) xs
    | _ -> raise (Generic_rdigest_failure "list_mem_nth")
  in
  sub 0

let debug0 s =
  let s = String.copy s in
  for i = 0 to String.length s - 1 do
    if s.[i] <> '\t' && (s.[i] < ' ' || s.[i] > '~') then s.[i] <- '.'
  done;
  prerr_endline ("\t"^s)

let debug s =
  try
    if Sys.getenv "DEBUG_DIGEST" <> "" then begin
      debug0 s
    end
  with
    Not_found -> ()

(*****************************************************************************)

(* built-in types *)
type rpath =
    RPident of string * int
  | RPdot of rpath * string * int
  | RPapply of rpath * rpath

type 'a rtype = 
    RTvar of int
  | RTarrow of 'a rtype * 'a rtype
  | RTtuple of 'a rtype list
  | RTconstr of 'a * 'a rtype list

type digest_info =
    Info_Digest of string (* = Digest.t *)
  | Info_Manifest of string rtype
  | Info_None

type lazy_digest_info_cell =
  | Value  of digest_info
  | Delayed of (unit -> digest_info)

type type_info = {
    path: rpath;
    mutable digest: lazy_digest_info_cell
  } 

let digest_info type_info =
  (* prerr_endline "digest_info"; *)
  let v = 
    match type_info.digest with
    | Value v -> v
    | Delayed f ->
	prerr_endline "delayed...";
	begin try
	  let v = f () in 
	  type_info.digest <- Value v; v
	with
	  e -> prerr_endline "digest_info failed"; raise e
	end
  in
  if v = Info_None then raise No_digest_info else v

type flags =
    Flag_Digest of string
  | Flag_Regular of int
  | Flag_Recursive of int

(* Type Digests *)

type 'a source_declaration =
    SType_abstract
  | SType_digest of string
  | SType_manifest of 'a rtype
  | SType_variant of 'a rtype list list
  | SType_record of 'a rtype list
  | SType_none
let catenate_sep sep =
  function 
      [] -> ""
    | x::l -> List.fold_left (fun s s' -> s^sep^s') x l

let rec string_of_rpath rp = 
  match rp with
    RPident (s,0) -> s
  | RPident (s,x) -> s ^ "/" ^ string_of_int x 
  | RPdot (p,s,0) -> string_of_rpath p ^ "." ^ s
  | RPdot (p,s,i) -> string_of_rpath p ^ "." ^ s^"/"^string_of_int i
  | RPapply (p1,p2) -> string_of_rpath p1 ^"("^ string_of_rpath p2 ^")"

let string_of_rtype f = 
  let rec sub = function
      RTvar x -> "'" ^ string_of_int x
    | RTarrow (f,t) -> "(" ^ sub f ^ " -> " ^ sub t ^ ")"
    | RTtuple typs -> "(* " ^ catenate_sep " * " (List.map sub typs) ^ ")" 
    | RTconstr (v, typs) ->
	if typs <> [] then 
	  "(" ^ catenate_sep ", " (List.map sub typs) ^ ") " ^ f v
	else f v
  in
  sub

let rec string_of_sdecl f = function
    SType_abstract -> raise (Generic_rdigest_failure "string_of_sdecl : abstract")
  | SType_none ->  raise (Generic_rdigest_failure "string_of_sdecl : none")
  | SType_digest s -> s
  | SType_manifest _ -> raise (Generic_rdigest_failure "string_of_sdecl : manifest")
  | SType_variant typslist ->
      catenate_sep " | " (List.map (fun typs -> 
	catenate_sep " * " (List.map (string_of_rtype f) typs)) 
			    typslist)
  | SType_record typs ->
      "{"^ catenate_sep "; " (List.map (string_of_rtype f) typs) ^"}"

let rec iter_rtype v c = function
    RTvar x -> v x
  | RTarrow (t1,t2) -> iter_rtype v c t1; iter_rtype v c t2
  | RTtuple typs -> List.iter (iter_rtype v c) typs
  | RTconstr (ci, typs) -> List.iter (iter_rtype v c) typs; c ci typs

let rec map_rtype v c = function
    RTvar x -> v x
  | RTarrow (t1,t2) -> RTarrow (map_rtype v c t1, map_rtype v c t2)
  | RTtuple typs -> RTtuple (List.map (map_rtype v c) typs)
  | RTconstr (ci, typs) -> c ci (List.map (map_rtype v c) typs)

let subst_rtype_at_var args typ =
  map_rtype (fun x -> 
    try
      List.nth args x
    with
      _ -> raise (Generic_rdigest_failure ("at_var nth at " ^ string_of_int x))) (fun ci typs -> RTconstr (ci, typs)) typ

let subst_rtype_at_flag_regular args typ =
  map_rtype (fun x -> RTvar x) (fun ci typs ->
    match ci with
      Flag_Regular x -> 
	begin match List.nth args x with
	  Info_Digest d -> RTconstr (d, typs)
	| Info_Manifest t -> subst_rtype_at_var typs t
	| Info_None -> raise (Generic_rdigest_failure "subst_rtype_at_flag_regular: None")
	end
    | Flag_Digest d -> RTconstr (d, typs)
    | Flag_Recursive _ -> 
	raise (Generic_rdigest_failure "subst_rtype_flag_regular: flag_recursive")) typ

let subst_rtype_at_flag_regular2 args typ =
  map_rtype (fun x -> RTvar x) (fun ci typs ->
    match ci with
      Flag_Regular x -> 
	begin match List.nth args x with
	  Info_Digest d -> RTconstr (Flag_Digest d, typs)
	| Info_Manifest t -> 
	    subst_rtype_at_var typs (map_rtype (fun x -> RTvar x) 
	       (fun ci typs -> RTconstr (Flag_Digest ci, typs)) t)
	| Info_None -> 
	    raise (Generic_rdigest_failure "subst_rtype_at_flag_regular2: none")
	end 
    | Flag_Digest d -> RTconstr (Flag_Digest d, typs)
    | Flag_Recursive x -> RTconstr (Flag_Recursive x, typs)) typ

let subst_sdecl args = function
    SType_abstract -> SType_abstract
  | SType_digest s -> SType_digest s
  | SType_manifest _ -> raise (Generic_rdigest_failure "subst_sdecl : manifest")
  | SType_variant typslist ->
      SType_variant (List.map (List.map (subst_rtype_at_flag_regular2 args)) typslist)
  | SType_record typs ->
      SType_record (List.map (subst_rtype_at_flag_regular2 args) typs)
  | SType_none -> SType_none

let manifest id rtyp args =
  debug ("Rdigest.manifest " ^ id ^ " = " ^
	 string_of_rtype (function
	     Flag_Digest d -> d
	   | Flag_Regular x -> "+" ^ string_of_int x
	   | Flag_Recursive x -> "#" ^ string_of_int x) rtyp
	 ^ " args=" ^ string_of_int (List.length args)  );
  List.iter (fun x -> debug ("args: " ^ 
     string_of_rpath x.path ^ " = " ^
      match x.digest with
       Delayed _ -> "Delayed"
     | Value _ -> "Value")) args;
  try
    let args = List.map digest_info args in
    let srtyp = subst_rtype_at_flag_regular args rtyp in
    debug ("Rdigest: " ^ id ^ " = " ^ 
	   string_of_rtype (fun x -> x) srtyp);
    Info_Manifest srtyp
  with
    No_digest_info -> 
      debug ("Rdigest: " ^ id ^ " = None");
      Info_None

let digest id src args =
  debug0 ("Rdigest.digest: " ^ id);
  try
    let args = List.map digest_info args in
    let source_string = 
      catenate_sep " and " (List.map (fun sdecl ->
	let sdecl' = subst_sdecl args sdecl in
	string_of_sdecl (function 
	    Flag_Digest d -> d
	  | Flag_Regular _ -> raise (Generic_rdigest_failure "Rdigest.digest : Flag_Regular")
	  | Flag_Recursive x -> "#" ^ string_of_int x) sdecl') src)
    in
    let digest = digest_string source_string in
    debug ("Rdigest: " ^ id ^ " = " ^ digest);
    Info_Digest digest
  with
    No_digest_info -> 
      debug ("Rdigest: " ^ id ^ " = None");
      Info_None
