(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  Distributed only by permission.                   *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typecore.ml,v 1.52 1999/02/12 12:34:32 weis Exp $ *)

(* Typechecking for the core language *)

let debug = try ignore (Sys.getenv "DEBUG_TYPECORE"); true with _ -> false
let debug_dyn = try ignore (Sys.getenv "DEBUG_DYNAMIC"); true with _ -> false

open Misc
open Asttypes
open Parsetree
open Types
open Typedtree
open Btype
open Ctype

type error =
    Unbound_value of Longident.t
  | Unbound_constructor of Longident.t
  | Unbound_label of Longident.t
  | Constructor_arity_mismatch of Longident.t * int * int
  | Label_mismatch of Longident.t * (type_expr * type_expr) list
  | Pattern_type_clash of (type_expr * type_expr) list
  | Multiply_bound_variable
  | Orpat_not_closed
  | Expr_type_clash of (type_expr * type_expr) list
  | Apply_non_function of type_expr
  | Label_multiply_defined of Longident.t
  | Label_missing
  | Label_not_mutable of Longident.t
  | Bad_format of string
  | Undefined_method of type_expr * string
  | Undefined_inherited_method of string
  | Unbound_class of Longident.t
  | Virtual_class of Longident.t
  | Unbound_instance_variable of string
  | Instance_variable_not_mutable of string
  | Not_subtype of (type_expr * type_expr) list * (type_expr * type_expr) list
  | Outside_class
  | Value_multiply_overridden of string
  | Coercion_failure of type_expr * type_expr * (type_expr * type_expr) list
  | Too_many_arguments
  | Scoping_let_module of string * type_expr
  | Masked_instance_variable of Longident.t

  | Generic_not_closed of type_expr
  | Generic_illegal_approximation of type_expr * type_expr
  | Generic_illegal_pattern
  | Generic_not_generic of type_expr
  | Generic_type_case_mismatch of type_expr * type_expr
  | Generic_type_case_not_generalizable of type_expr
  | Coerce_pattern_needs_type_info of type_expr * type_expr

(* KONST
  | No_match_konstraint of Path.t * type_expr
  | Ambiguous_konstraints of (Path.t * type_expr) list
*)

exception Error of Location.t * error

(* KONST
let gkonst_typing gids =
  if gids = [] then [] 
  else
    try
      Gkonst.konstraint_typing gids
    with
    | Gkonst.No_match_konstraint (texp, path, typ) ->
      	raise (Error (texp.exp_loc, No_match_konstraint (path, typ)))
    | Gkonst.Ambiguous_konstraints (texp, pathtyps) ->
      	raise (Error (texp.exp_loc, Ambiguous_konstraints pathtyps))
*)

(* Forward declaration, to be filled in by Typemod.type_module *)

let type_module =
  ref ((fun env md -> assert false) :
       Env.t -> Parsetree.module_expr -> Typedtree.module_expr)

(* Typing of constants *)

let type_constant = function
    Const_int _ -> instance Predef.type_int
  | Const_char _ -> instance Predef.type_char
  | Const_string _ -> instance Predef.type_string
  | Const_float _ -> instance Predef.type_float

(* Typing of patterns *)

let unify_pat env pat expected_ty =
  try
    unify env pat.pat_type expected_ty
  with Unify trace ->
    raise(Error(pat.pat_loc, Pattern_type_clash(trace)))

let pattern_variables = ref ([]: (Ident.t * type_expr) list)

let enter_variable loc name ty =
  if List.exists (fun (id, ty) -> Ident.name id = name) !pattern_variables
  then raise(Error(loc, Multiply_bound_variable));
  let id = Ident.create name in
  pattern_variables := (id, ty) :: !pattern_variables;
  id

let type_pat env sp =
  let explicit_type_vars = ref [] in
  let any_type_vars = ref [] in
  let rec type_pat env sp =
    match sp.ppat_desc with
      Ppat_any ->
	let var = newvar () in
	if !Ctype.inside_coerce_typecase then 
	  any_type_vars := var :: !any_type_vars;
	{ pat_desc = Tpat_any (new_generic_info ());
  	  pat_loc = sp.ppat_loc;
  	  pat_type = var;
  	  pat_env = env }
    | Ppat_var name ->
  	let ty = newvar() in
  	let id = enter_variable sp.ppat_loc name ty in
  	{ pat_desc = Tpat_var (id, (new_generic_info ()));
  	  pat_loc = sp.ppat_loc;
  	  pat_type = ty;
  	  pat_env = env }
    | Ppat_alias(sp, name) ->
  	let p = type_pat env sp in
  	let id = enter_variable sp.ppat_loc name p.pat_type in
  	{ pat_desc = Tpat_alias(p, id);
  	  pat_loc = sp.ppat_loc;
  	  pat_type = p.pat_type;
  	  pat_env = env }
    | Ppat_constant cst ->
  	{ pat_desc = Tpat_constant cst;
  	  pat_loc = sp.ppat_loc;
  	  pat_type = type_constant cst;
  	  pat_env = env }
    | Ppat_tuple spl ->
  	let pl = List.map (type_pat env) spl in
  	{ pat_desc = Tpat_tuple pl;
  	  pat_loc = sp.ppat_loc;
  	  pat_type = newty (Ttuple(List.map (fun p -> p.pat_type) pl));
  	  pat_env = env }
    | Ppat_construct(lid, sarg, explicit_arity) ->
  	let constr =
  	  try
  	    Env.lookup_constructor lid env
  	  with Not_found ->
  	    raise(Error(sp.ppat_loc, Unbound_constructor lid)) in
  	let sargs =
  	  match sarg with
  	    None -> []
  	  | Some {ppat_desc = Ppat_tuple spl} when explicit_arity -> spl
  	  | Some {ppat_desc = Ppat_tuple spl} when constr.cstr_arity > 1 -> spl
  	  | Some({ppat_desc = Ppat_any} as sp) when constr.cstr_arity > 1 ->
  	      replicate_list sp constr.cstr_arity
  	  | Some sp -> [sp] in
  	if List.length sargs <> constr.cstr_arity then
  	  raise(Error(sp.ppat_loc, Constructor_arity_mismatch(lid,
  				       constr.cstr_arity, List.length sargs)));
  	let args = List.map (type_pat env) sargs in
  	let (ty_args, ty_res) = instance_constructor constr in
  	List.iter2 (unify_pat env) args ty_args;
  	{ pat_desc = Tpat_construct(constr, args);
  	  pat_loc = sp.ppat_loc;
  	  pat_type = ty_res;
  	  pat_env = env }
    | Ppat_record lid_sp_list ->
  	let rec check_duplicates = function
  	  [] -> ()
  	| (lid, sarg) :: remainder ->
  	    if List.mem_assoc lid remainder
  	    then raise(Error(sp.ppat_loc, Label_multiply_defined lid))
  	    else check_duplicates remainder in
  	check_duplicates lid_sp_list;
  	let ty = newvar() in
  	let type_label_pat (lid, sarg) =
  	  let label =
  	    try
  	      Env.lookup_label lid env
  	    with Not_found ->
  	      raise(Error(sp.ppat_loc, Unbound_label lid)) in
  	  let (ty_arg, ty_res) = instance_label label in
  	  begin try
  	    unify env ty_res ty
  	  with Unify trace ->
  	    raise(Error(sp.ppat_loc, Label_mismatch(lid, trace)))
  	  end;
  	  let arg = type_pat env sarg in
  	  unify_pat env arg ty_arg;
  	  (label, arg)
  	in
  	{ pat_desc = Tpat_record(List.map type_label_pat lid_sp_list);
  	  pat_loc = sp.ppat_loc;
  	  pat_type = ty;
  	  pat_env = env }
    | Ppat_array spl ->
  	let pl = List.map (type_pat env) spl in
  	let ty_elt = newvar() in
  	List.iter (fun p -> unify_pat env p ty_elt) pl;
  	{ pat_desc = Tpat_array pl;
  	  pat_loc = sp.ppat_loc;
  	  pat_type = instance (Predef.type_array ty_elt);
  	  pat_env = env }
    | Ppat_or(sp1, sp2) ->
  	let initial_pattern_variables = !pattern_variables in
  	let p1 = type_pat env sp1 in
  	let p2 = type_pat env sp2 in
  	if !pattern_variables != initial_pattern_variables then
  	  raise(Error(sp.ppat_loc, Orpat_not_closed));
  	unify_pat env p2 p1.pat_type;
  	{ pat_desc = Tpat_or(p1, p2);
  	  pat_loc = sp.ppat_loc;
  	  pat_type = p1.pat_type;
  	  pat_env = env }
    | Ppat_constraint(sp, sty) ->
  	let p = type_pat env sp in
  	if !Ctype.inside_coerce_typecase then begin
	  let ty = Typetexp.transl_pattern_type_scheme env sty in
	  let ty = instance ty in (* anyvars are shared *)
	  (* extract tvars in ty and store them *)
	  let nvars = Dtype.normal_variables ty in
	  List.iter (fun nvar ->
	    if not (List.memq nvar !explicit_type_vars) then
	      explicit_type_vars := nvar :: !explicit_type_vars) nvars;
	  (* inside coerce typecase, p.pat_type > ty *)
	  if not (moregeneral env true p.pat_type ty) then begin
	    raise (Error (p.pat_loc, Generic_type_case_mismatch (p.pat_type, ty)))
	  end;
	end else begin
	  let ty = Typetexp.transl_simple_type env false sty in
	  unify_pat env p ty
	end;
  	p
  in
  if !Ctype.inside_coerce_typecase then begin
    let p = type_pat env sp in
    List.iter (fun (i,t) ->
      let nvars = Dtype.normal_variables t in
      List.iter (fun nvar ->
	if not (List.memq nvar !explicit_type_vars) then
	  raise (Error (p.pat_loc, 
			Coerce_pattern_needs_type_info (p.pat_type, nvar)))) 
	nvars) !pattern_variables;
    p
  end else type_pat env sp

let is_extensible = ref false (* used for type_generic_pattern_list *)

let add_pattern_variables env =
  let pv = !pattern_variables in
  pattern_variables := [];
  List.fold_right
    (fun (id, ty) env ->
       Env.add_value id { val_type = ty; 
			  val_kind = if !is_extensible then Val_extensible
			             else Val_reg;
			  val_konstraints = Konst_None
			} env)
    pv env

let type_pattern env spat =
  pattern_variables := [];
  let pat = type_pat env spat in
  let new_env = add_pattern_variables env in
  (pat, new_env)

let type_pattern_list env spatl =
  pattern_variables := [];
  let patl = List.map (type_pat env) spatl in
  let new_env = add_pattern_variables env in
  (patl, new_env)

let generic_moregeneral env gscm pscm = (* check gscm > pscm *) 
  if debug then begin
    Format.print_string "    Check gscm > pscm : ";
    Printtyp.type_scheme gscm;
    Format.print_string " ?>? ";
    Printtyp.type_scheme pscm;
    Format.print_newline ()
  end;

  (* we need to instance gscm and pscm in this level *)
  (* check gscm > pscm *)
  let gscm' = instance gscm in
  let pscm' = instance pscm in
  (* dvars_scm' must be got before any unification *)
  let dvars_scm' = Dtype.dynamic_variables gscm' in

  if debug then begin
    Format.print_string "      Check gscm' > pscm' : ";
    Printtyp.type_scheme gscm';
    Format.print_string " ?>? ";
    Printtyp.type_scheme pscm';
    Format.print_newline ()
  end;

  (* we check gscm > pscm by gscm' > pscm' because we want to 
     keep the originals *)
  if not (moregeneral env true gscm' pscm') then begin
    raise Exit
  end;

  if debug then begin
    Format.print_string "      Ok, gscm > pscm : ";
    Printtyp.type_scheme gscm;
    Format.print_string " > ";
    Printtyp.type_scheme pscm;
    Format.print_newline ();
(*
    Format.print_string "  Returns ";
    Printtyp.type_scheme pscm';
    Format.print_string " [ ";
    List.iter (fun typ -> 
      Printtyp.type_scheme typ;
      Format.print_string "; ") dvars_scm';
    Format.print_string " ]";
    Format.print_newline ();
*)
  end;
  
  pscm', dvars_scm'
;;

let type_generic_pattern_list env extensible_flag bindings =

  (* pattern_variables := []; *)
  let generic_pattern_variables = ref [] in

  let generic_enter_variable loc name ty =
    if List.exists (fun (id, ty) -> Ident.name id = name) 
	!generic_pattern_variables
    then raise(Error(loc, Multiply_bound_variable));
    let id = Ident.create name in
    generic_pattern_variables := 
       (id, ty) :: !generic_pattern_variables;
    id
  in

  let generic_add_pattern_variables env =
    let pv = !generic_pattern_variables in
    generic_pattern_variables := [];
    List.fold_right
      (fun (id, ty) env ->
       	Env.add_value id { val_type = ty; 
			   val_kind = if !is_extensible then Val_extensible
			              else Val_reg;
			   val_konstraints = Konst_None
			 } env)
      pv env
  in

  (* unlike let, patterns are polymorphic inside their recursive defs *) 
  let patl = List.map (fun (spat,scases) ->
    begin_def ();

    let name, tyopt, styopt =
      match spat.ppat_desc with
      | Ppat_var name -> 
	  name, None, None
      |	Ppat_constraint ({ppat_desc= Ppat_var name}, sty) ->
	  name, Some (Typetexp.transl_simple_type env false sty), Some sty
      | _ -> 
	  (* just reject everything instead of
	       generic t = 
	         or
	       generic (t : typ) = *)
	  raise (Error (spat.ppat_loc, Generic_illegal_pattern))
    in
    
    let pattern_schemes =
      List.map (function
	| Pgeneric_typecase (sty, _) ->
	    Typetexp.transl_pattern_type_scheme env sty
	| Pgeneric_include (lid,loc) -> 
	    (* we do the same again... *)
	    let (path,desc) = Env.lookup_value lid env in
	    desc.val_type) scases
    in

    let aty = (* anti-unified type scheme of patterns *)
      Dtype.anti_unify_types env pattern_schemes
    in
    end_def ();

    let ty = 
      (* get the generalized approximation scheme *)
      (* here we do not need to copy generalize 
	 because no links exist *)
      generalize aty;
      match tyopt, styopt with
      |	Some ty, Some sty -> 
	  generalize ty;
	  begin try
	    ignore (generic_moregeneral env ty aty);
	    ty
	  with
	    Exit -> 
	      raise (Error (sty.ptyp_loc, 
			    Generic_illegal_approximation (ty,aty)))
	  end
      |	None, None -> aty
      |	_ -> assert false
    in

    (* ty > aty *)
(* this is not correct
    unify env aty ty;
*)

(*
    if debug then begin
      Printtyp.type_scheme ty;
      Format.print_newline ();
    end;
*)

    if not (Ctype.closed_schema ty) then 
      raise (Error (spat.ppat_loc, Generic_not_closed ty));

    if Dtype.dynamic_variables ty = [] then 
      raise (Error (spat.ppat_loc, Generic_not_generic ty));

    let id = generic_enter_variable spat.ppat_loc name ty in
    { pat_desc= Tpat_var (id, (new_generic_info ()));
      pat_loc= spat.ppat_loc;
      pat_type= ty;
      pat_env= env }) bindings
  in
  try
    is_extensible := (extensible_flag != Nonextensible);
    let new_env = generic_add_pattern_variables env in
    is_extensible := false;
    (patl, new_env)
  with
    e -> is_extensible := false; raise e

let type_class_arg_pattern cl_num val_env met_env spat =
  pattern_variables := [];
  let pat = type_pat val_env spat in
  let (pv, met_env) =
    List.fold_right
      (fun (id, ty) (pv, env) ->
         let id' = Ident.create (Ident.name id) in
         ((id', id, ty)::pv,
          Env.add_value id' {val_type = ty;
                             val_kind = Val_ivar (Immutable, cl_num);
			     val_konstraints = Konst_None}
            env))
      !pattern_variables ([], met_env)
  in
  let val_env = add_pattern_variables val_env in
  (pat, pv, val_env, met_env)

let mkpat d = { ppat_desc = d; ppat_loc = Location.none }
let type_self_pattern cl_num val_env met_env par_env spat =
  let spat = 
    mkpat (Ppat_alias (mkpat(Ppat_alias (spat, "selfpat-*")),
                       "selfpat-" ^ cl_num))
  in
  pattern_variables := [];
  let pat = type_pat val_env spat in
  let meths = ref Meths.empty in
  let vars = ref Vars.empty in
  let pv = !pattern_variables in
  pattern_variables := [];
  let (val_env, met_env, par_env) =
    List.fold_right
      (fun (id, ty) (val_env, met_env, par_env) ->
         (Env.add_value id {val_type = ty; val_kind = Val_unbound;
			    val_konstraints = Konst_None} val_env,
          Env.add_value id {val_type = ty;
                            val_kind = Val_self (meths, vars, cl_num);
			    val_konstraints = Konst_None}
            met_env,
          Env.add_value id {val_type = ty; val_kind = Val_unbound;
			    val_konstraints = Konst_None} par_env))
      pv (val_env, met_env, par_env)
  in
  (pat, meths, vars, val_env, met_env, par_env)

let rec iter_pattern f p =
  f p;
  match p.pat_desc with
    Tpat_any _ | Tpat_var (_,_) | Tpat_constant _ ->
      ()
  | Tpat_alias (p, _) ->
      iter_pattern f p
  | Tpat_tuple pl ->
      List.iter (iter_pattern f) pl
  | Tpat_construct (_, pl) ->
      List.iter (iter_pattern f) pl
  | Tpat_record fl ->
      List.iter (fun (_, p) -> iter_pattern f p) fl
  | Tpat_or (p, p') ->
      iter_pattern f p;
      iter_pattern f p'
  | Tpat_array pl ->
      List.iter (iter_pattern f) pl

(* Generalization criterion for expressions *)

let rec is_nonexpansive exp =
  match exp.exp_desc with
    Texp_ident(_,_,_) -> true
  | Texp_constant _ -> true
  | Texp_let(rec_flag, pat_exp_list, body) ->
      List.for_all (fun (pat, exp) -> is_nonexpansive exp) pat_exp_list &
      is_nonexpansive body
  | Texp_generic(_, _, _, body) -> is_nonexpansive body
  | Texp_function _ -> true
  | Texp_tuple el ->
      List.for_all is_nonexpansive el
  | Texp_construct(_, el) ->
      List.for_all is_nonexpansive el
  | Texp_record(lbl_exp_list, opt_init_exp) ->
      List.for_all
        (fun (lbl, exp) -> lbl.lbl_mut = Immutable & is_nonexpansive exp)
        lbl_exp_list &&
      (match opt_init_exp with None -> true | Some e -> is_nonexpansive e)
  | Texp_field(exp, lbl) -> is_nonexpansive exp
  | Texp_array [] -> true
  | Texp_new (_, cl_decl) when Ctype.class_type_arity cl_decl.cty_type > 0 ->
      true
  | _ -> false

(* Typing of printf formats *)

let type_format loc fmt =
  let len = String.length fmt in
  let ty_input = newvar()
  and ty_result = newvar() in
  let rec skip_args j =
    if j >= len then j else
      match fmt.[j] with
        '0' .. '9' | ' ' | '.' | '-' -> skip_args (j+1)
      | _ -> j in
  let rec scan_format i =
    if i >= len then ty_result else
    match fmt.[i] with
      '%' ->
        let j = skip_args(i+1) in
        if j >= len then raise(Error(loc, Bad_format "%"));
        begin match fmt.[j] with
          '%' ->
            scan_format (j+1)
        | 's' ->
            newty (Tarrow(instance Predef.type_string, scan_format (j+1)))
        | 'c' ->
            newty (Tarrow(instance Predef.type_char, scan_format (j+1)))
        | 'd' | 'i' | 'o' | 'x' | 'X' | 'u' ->
            newty (Tarrow(instance Predef.type_int, scan_format (j+1)))
        | 'f' | 'e' | 'E' | 'g' | 'G' ->
            newty (Tarrow(instance Predef.type_float, scan_format (j+1)))
        | 'b' ->
            newty (Tarrow(instance Predef.type_bool, scan_format (j+1)))
        | 'a' ->
            let ty_arg = newvar() in
            newty (Tarrow (newty (Tarrow(ty_input,
                                         newty (Tarrow (ty_arg, ty_result)))),
                           newty (Tarrow (ty_arg, scan_format (j+1)))))
        | 't' ->
            newty (Tarrow(newty (Tarrow(ty_input, ty_result)),
                          scan_format (j+1)))
        | c ->
            raise(Error(loc, Bad_format(String.sub fmt i (j-i+1))))
        end
    | _ -> scan_format (i+1) in
  newty
    (Tconstr(Predef.path_format, [scan_format 0; ty_input; ty_result], ref Mnil))

(* Typing of expressions *)

let unify_exp env exp expected_ty =
  try
    unify env exp.exp_type expected_ty
  with Unify trace ->
    raise(Error(exp.exp_loc, Expr_type_clash(trace)))

let rec type_exp env sexp =
  match sexp.pexp_desc with
    Pexp_ident lid ->
      begin try
        let (path, desc) = Env.lookup_value lid env in
	let texp = 
        { exp_desc =
            begin match desc.val_kind with
              Val_ivar (_, cl_num) ->
                let (self_path, _) =
                  Env.lookup_value (Longident.Lident ("self-" ^ cl_num)) env
                in
                Texp_instvar(self_path, path)
            | Val_self (_, _, cl_num) ->
                let (path, _) =
                  Env.lookup_value (Longident.Lident ("self-" ^ cl_num)) env
                in
                Texp_ident(path, desc, ref Inst_normal)
            | Val_unbound ->
                raise(Error(sexp.pexp_loc, Masked_instance_variable lid))
            | _ ->
                Texp_ident(path, desc, ref Inst_normal)
            end;
          exp_loc = sexp.pexp_loc;
          exp_type = instance desc.val_type;
          exp_env = env }
	in
	texp
      with Not_found ->
        raise(Error(sexp.pexp_loc, Unbound_value lid))
      end
  | Pexp_constant cst ->
      { exp_desc = Texp_constant cst;
        exp_loc = sexp.pexp_loc;
        exp_type = type_constant cst;
        exp_env = env }
  | Pexp_let(rec_flag, spat_sexp_list, sbody) ->
      let (pat_exp_list, new_env) = type_let env rec_flag spat_sexp_list in
      let body = type_exp new_env sbody in
      { exp_desc = Texp_let(rec_flag, pat_exp_list, body);
        exp_loc = sexp.pexp_loc;
        exp_type = body.exp_type;
        exp_env = env }
  | Pexp_function caselist ->
      let ty_arg = newvar() and ty_res = newvar() in
      let cases = type_cases env ty_arg ty_res caselist in
      Parmatch.check_unused cases;
      Parmatch.check_partial sexp.pexp_loc cases;
      { exp_desc = Texp_function cases;
        exp_loc = sexp.pexp_loc;
        exp_type = newty (Tarrow(ty_arg, ty_res));
        exp_env = env }
  | Pexp_apply(sfunct, sargs) ->
      let need_dynamic_polymorphism =
	match sfunct.pexp_desc with
	  Pexp_ident lid ->
	    let _, desc = 
	      try
		Env.lookup_value lid env 
	      with Not_found ->
		raise(Error(sfunct.pexp_loc, Unbound_value lid))
	    in
	    Dtype.generalized_dynamic_variables desc.val_type <> []
	| _ -> false
      in
      if need_dynamic_polymorphism then begin
	begin_def ();
      end;

      let funct = type_exp env sfunct in
      let rec type_args ty_fun = function
        [] ->
          ([], ty_fun)
      | sarg1 :: sargl ->
          let (ty1, ty2) =
            try
              filter_arrow env ty_fun
            with Unify _ ->
              raise(Error(sfunct.pexp_loc,
                          Apply_non_function funct.exp_type)) in
          let arg1 = type_expect env sarg1 ty1 in
          let (argl, ty_res) = type_args ty2 sargl in
          (arg1 :: argl, ty_res) in
      let (args, ty_res) = type_args funct.exp_type sargs in

      if need_dynamic_polymorphism then begin
	end_def ();
	List.iter (fun arg -> 
	  ignore (Dtype.generalized_variables ()); (* reset the table *)
	  if is_nonexpansive arg then ignore (Dtype.generalize_copy arg.exp_type);
	  List.iter (fun (org,gen) ->
	    match org.desc with
	    | Tdvar (x,y) -> 
		(* reset the generalization info
		   and put Exportable instead *)
		x := Not_generalized; y := Exportable
	    | _ -> assert false) (Dtype.generalized_variables ());
	  
	if debug_dyn then begin
	  Format.print_string "dyn arg: ";
	  Printtyp.type_expr arg.exp_type;
	  Format.print_newline ()
	end ) args;
	
	if debug_dyn then begin
	  Format.print_string "dyn func: ";
	  Printtyp.type_expr funct.exp_type;
	  Format.print_newline ()
	end
      end;

      { exp_desc = Texp_apply(funct, args);
        exp_loc = sexp.pexp_loc;
        exp_type = ty_res;
        exp_env = env }
  | Pexp_match(sarg, caselist) ->
      let arg = type_exp env sarg in
      let ty_res = newvar() in
      let cases = type_cases env arg.exp_type ty_res caselist in
      Parmatch.check_unused cases;
      Parmatch.check_partial sexp.pexp_loc cases;
      { exp_desc = Texp_match(arg, cases);
        exp_loc = sexp.pexp_loc;
        exp_type = ty_res;
        exp_env = env }
  | Pexp_try(sbody, caselist) ->
      let body = type_exp env sbody in
      let cases =
        type_cases env (instance Predef.type_exn) body.exp_type caselist in
      Parmatch.check_unused cases;
      { exp_desc = Texp_try(body, cases);
        exp_loc = sexp.pexp_loc;
        exp_type = body.exp_type;
        exp_env = env }
  | Pexp_tuple sexpl ->
      let expl = List.map (type_exp env) sexpl in
      { exp_desc = Texp_tuple expl;
        exp_loc = sexp.pexp_loc;
        exp_type = newty (Ttuple(List.map (fun exp -> exp.exp_type) expl));
        exp_env = env }
  | Pexp_construct(lid, sarg, explicit_arity) ->
      let constr =
        try
          Env.lookup_constructor lid env
        with Not_found ->
          raise(Error(sexp.pexp_loc, Unbound_constructor lid)) in
      let sargs =
        match sarg with
          None -> []
        | Some {pexp_desc = Pexp_tuple sel} when explicit_arity -> sel
        | Some {pexp_desc = Pexp_tuple sel} when constr.cstr_arity > 1 -> sel
        | Some se -> [se] in
      if List.length sargs <> constr.cstr_arity then
        raise(Error(sexp.pexp_loc, Constructor_arity_mismatch(lid,
                                       constr.cstr_arity, List.length sargs)));
      let (ty_args, ty_res) = instance_constructor constr in
      let args = List.map2 (type_expect env) sargs ty_args in
      { exp_desc = Texp_construct(constr, args);
        exp_loc = sexp.pexp_loc;
        exp_type = ty_res;
        exp_env = env }
  | Pexp_record(lid_sexp_list, opt_sexp) ->
      let ty = newvar() in
      let num_fields = ref 0 in
      let type_label_exp (lid, sarg) =
        let label =
          try
            Env.lookup_label lid env
          with Not_found ->
            raise(Error(sexp.pexp_loc, Unbound_label lid)) in
        let (ty_arg, ty_res) = instance_label label in
        begin try
          unify env ty_res ty
        with Unify trace ->
          raise(Error(sexp.pexp_loc, Label_mismatch(lid, trace)))
        end;
        let arg = type_expect env sarg ty_arg in
        num_fields := Array.length label.lbl_all;
        (label, arg) in
      let lbl_exp_list = List.map type_label_exp lid_sexp_list in
      let rec check_duplicates = function
        [] -> ()
      | (lid, sarg) :: remainder ->
          if List.mem_assoc lid remainder
          then raise(Error(sexp.pexp_loc, Label_multiply_defined lid))
          else check_duplicates remainder in
      check_duplicates lid_sexp_list;
      let opt_exp =
        match opt_sexp with
          None -> None
        | Some sexp -> Some(type_expect env sexp ty) in
      if opt_sexp = None && List.length lid_sexp_list <> !num_fields then
        raise(Error(sexp.pexp_loc, Label_missing));
      { exp_desc = Texp_record(lbl_exp_list, opt_exp);
        exp_loc = sexp.pexp_loc;
        exp_type = ty;
        exp_env = env }
  | Pexp_field(sarg, lid) ->
      let arg = type_exp env sarg in
      let label =
        try
          Env.lookup_label lid env
        with Not_found ->
          raise(Error(sexp.pexp_loc, Unbound_label lid)) in
      let (ty_arg, ty_res) = instance_label label in
      unify_exp env arg ty_res;
      { exp_desc = Texp_field(arg, label);
        exp_loc = sexp.pexp_loc;
        exp_type = ty_arg;
        exp_env = env }
  | Pexp_setfield(srecord, lid, snewval) ->
      let record = type_exp env srecord in
      let label =
        try
          Env.lookup_label lid env
        with Not_found ->
          raise(Error(sexp.pexp_loc, Unbound_label lid)) in
      if label.lbl_mut = Immutable then
        raise(Error(sexp.pexp_loc, Label_not_mutable lid));
      let (ty_arg, ty_res) = instance_label label in
      unify_exp env record ty_res;
      let newval = type_expect env snewval ty_arg in
      { exp_desc = Texp_setfield(record, label, newval);
        exp_loc = sexp.pexp_loc;
        exp_type = instance Predef.type_unit;
        exp_env = env }
  | Pexp_array(sargl) ->
      let ty = newvar() in
      let argl = List.map (fun sarg -> type_expect env sarg ty) sargl in
      { exp_desc = Texp_array argl;
        exp_loc = sexp.pexp_loc;
        exp_type = instance (Predef.type_array ty);
        exp_env = env }
  | Pexp_ifthenelse(scond, sifso, sifnot) ->
      let cond = type_expect env scond (instance Predef.type_bool) in
      begin match sifnot with
        None ->
          let ifso = type_expect env sifso (instance Predef.type_unit) in
          { exp_desc = Texp_ifthenelse(cond, ifso, None);
            exp_loc = sexp.pexp_loc;
            exp_type = instance Predef.type_unit;
            exp_env = env }
      | Some sifnot ->
          let ifso = type_exp env sifso in
          let ifnot = type_expect env sifnot ifso.exp_type in
          { exp_desc = Texp_ifthenelse(cond, ifso, Some ifnot);
            exp_loc = sexp.pexp_loc;
            exp_type = ifso.exp_type;
            exp_env = env }
      end
  | Pexp_sequence(sexp1, sexp2) ->
      let exp1 = type_statement env sexp1 in
      let exp2 = type_exp env sexp2 in
      { exp_desc = Texp_sequence(exp1, exp2);
        exp_loc = sexp.pexp_loc;
        exp_type = exp2.exp_type;
        exp_env = env }
  | Pexp_while(scond, sbody) ->
      let cond = type_expect env scond (instance Predef.type_bool) in
      let body = type_statement env sbody in
      { exp_desc = Texp_while(cond, body);
        exp_loc = sexp.pexp_loc;
        exp_type = instance Predef.type_unit;
        exp_env = env }
  | Pexp_for(param, slow, shigh, dir, sbody) ->
      let low = type_expect env slow (instance Predef.type_int) in
      let high = type_expect env shigh (instance Predef.type_int) in
      let (id, new_env) =
        Env.enter_value param {val_type = instance Predef.type_int;
                                val_kind = Val_reg;
			      val_konstraints = Konst_None} env in
      let body = type_statement new_env sbody in
      { exp_desc = Texp_for(id, low, high, dir, body);
        exp_loc = sexp.pexp_loc;
        exp_type = instance Predef.type_unit;
        exp_env = env }
  | Pexp_constraint(sarg, sty, sty') ->
      let (arg, ty') =
        match (sty, sty') with
          (None, None) ->               (* Case actually unused *)
            let arg = type_exp env sarg in
            (arg, arg.exp_type)
        | (Some sty, None) ->
            let ty = Typetexp.transl_simple_type env false sty in
            (type_expect env sarg ty, ty)
        | (None, Some sty') ->
            let (ty', force) =
              Typetexp.transl_simple_type_delayed env sty'
            in
            let ty = enlarge_type env ty' in
            force ();
            let arg = type_exp env sarg in
            begin try Ctype.unify env arg.exp_type ty with Unify trace ->
              raise(Error(sarg.pexp_loc,
                    Coercion_failure(ty', full_expand env ty', trace)))
            end;
            (arg, ty')
        | (Some sty, Some sty') ->
            let (ty, force) =
              Typetexp.transl_simple_type_delayed env sty
            and (ty', force') =
              Typetexp.transl_simple_type_delayed env sty'
            in
            begin try
              let force'' = subtype env ty ty' in
              force (); force' (); force'' ()
            with Subtype (tr1, tr2) ->
              raise(Error(sexp.pexp_loc, Not_subtype(tr1, tr2)))
            end;
            (type_expect env sarg ty, ty')
      in
      { exp_desc = arg.exp_desc;
        exp_loc = arg.exp_loc;
        exp_type = ty';
        exp_env = env }
  | Pexp_when(scond, sbody) ->
      let cond = type_expect env scond (instance Predef.type_bool) in
      let body = type_exp env sbody in
      { exp_desc = Texp_when(cond, body);
        exp_loc = sexp.pexp_loc;
        exp_type = body.exp_type;
        exp_env = env }
  | Pexp_send (e, met) ->
      let obj = type_exp env e in
      begin try
        let (exp, typ) =
          match obj.exp_desc with
            Texp_ident(path, {val_kind = Val_self (meths, _, _)},_) ->
              let (id, typ) =
                filter_self_method env met Private meths obj.exp_type
              in
              (Texp_send(obj, Tmeth_val id), typ)
          | Texp_ident(path, {val_kind = Val_anc (methods, cl_num)},_) ->
              let method_id =
                begin try List.assoc met methods with Not_found ->
                  raise(Error(e.pexp_loc, Undefined_inherited_method met))
                end
              in
              begin match
                Env.lookup_value (Longident.Lident ("selfpat-" ^ cl_num)) env,
                Env.lookup_value (Longident.Lident ("self-" ^cl_num)) env
              with
                (_, ({val_kind = Val_self (meths, _, _)} as desc)),
                (path, _) ->
                  let (_, typ) =
                    filter_self_method env met Private meths obj.exp_type
                  in
                  let method_type = newvar () in
                  let (obj_ty, res_ty) = filter_arrow env method_type in
                  unify env obj_ty desc.val_type;
                  unify env res_ty typ;
                  (Texp_apply({exp_desc = Texp_ident(Path.Pident method_id,
                                                     {val_type = method_type;
                                                       val_kind = Val_reg;
						     val_konstraints = 
						       Konst_None},
						     ref Inst_normal);
                                exp_loc = sexp.pexp_loc;
                                exp_type = method_type;
                                exp_env = env },
                              [{exp_desc = Texp_ident(path, desc, ref Inst_normal);
                                 exp_loc = obj.exp_loc;
                                 exp_type = desc.val_type;
                                 exp_env = env }]),
                   typ)
              |  _ ->
                  assert false
              end
          | _ ->
              (Texp_send(obj, Tmeth_name met),
               filter_method env met Public obj.exp_type)
        in
          { exp_desc = exp;
            exp_loc = sexp.pexp_loc;
            exp_type = typ;
            exp_env = env }
      with Unify _ ->
        raise(Error(e.pexp_loc, Undefined_method (obj.exp_type, met)))
      end
  | Pexp_new cl ->
      let (cl_path, cl_decl) =
        try Env.lookup_class cl env with Not_found ->
          raise(Error(sexp.pexp_loc, Unbound_class cl))
      in
        begin match cl_decl.cty_new with
          None ->
            raise(Error(sexp.pexp_loc, Virtual_class cl))
        | Some ty ->
            { exp_desc = Texp_new (cl_path, cl_decl);
              exp_loc = sexp.pexp_loc;
              exp_type = instance ty;
              exp_env = env }
        end
  | Pexp_setinstvar (lab, snewval) ->
      begin try
        let (path, desc) = Env.lookup_value (Longident.Lident lab) env in
        match desc.val_kind with
          Val_ivar (Mutable, cl_num) ->
            let newval = type_expect env snewval desc.val_type in
            let (path_self, _) =
              Env.lookup_value (Longident.Lident ("self-" ^ cl_num)) env
            in
            { exp_desc = Texp_setinstvar(path_self, path, newval);
              exp_loc = sexp.pexp_loc;
              exp_type = instance Predef.type_unit;
              exp_env = env }
        | Val_ivar _ ->
            raise(Error(sexp.pexp_loc, Instance_variable_not_mutable lab))
        | _ ->
            raise(Error(sexp.pexp_loc, Unbound_instance_variable lab))
      with
        Not_found ->
          raise(Error(sexp.pexp_loc, Unbound_instance_variable lab))
      end        
  | Pexp_override lst ->
      let _ = 
       List.fold_right
        (fun (lab, _) l ->
           if List.exists ((=) lab) l then
             raise(Error(sexp.pexp_loc,
                         Value_multiply_overridden lab));
           lab::l)
        lst
        [] in
      begin match
        try
          Env.lookup_value (Longident.Lident "selfpat-*") env,
          Env.lookup_value (Longident.Lident "self-*") env
        with Not_found ->
          raise(Error(sexp.pexp_loc, Outside_class))
      with
        (_, {val_type = self_ty; val_kind = Val_self (_, vars, _)}),
        (path_self, _) ->
          let type_override (lab, snewval) =
            begin try
              let (id, _, ty) = Vars.find lab !vars in
              (Path.Pident id, type_expect env snewval ty)
            with
              Not_found ->
                raise(Error(sexp.pexp_loc, Unbound_instance_variable lab))
            end
          in
          let modifs = List.map type_override lst in
          { exp_desc = Texp_override(path_self, modifs);
            exp_loc = sexp.pexp_loc;
            exp_type = self_ty;
            exp_env = env }
      | _ ->
          assert false
      end
  | Pexp_letmodule(name, smodl, sbody) ->
      let ty = newvar() in
      Ident.set_current_time ty.level;
      let modl = !type_module env smodl in
      let (id, new_env) = Env.enter_module name modl.mod_type env in
      Ctype.init_def(Ident.current_time());
      let body = type_exp new_env sbody in
      (* Unification of body.exp_type with the fresh variable ty
         fails if and only if the prefix condition is violated,
         i.e. if generative types rooted at id show up in the
         type body.exp_type.  Thus, this unification enforces the
         scoping condition on "let module". *)
      begin try
        Ctype.unify new_env body.exp_type ty
      with Unify _ ->
        raise(Error(sexp.pexp_loc, Scoping_let_module(name, body.exp_type)))
      end;
      { exp_desc = Texp_letmodule(id, modl, body);
        exp_loc = sexp.pexp_loc;
        exp_type = ty;
        exp_env = env }
(*
  | Pexp_typeinfo type_longident ->
      let (path,decl) =
      try
	Env.lookup_type type_longident env
      with Not_found ->
	raise (Typetexp.Error(sexp.pexp_loc, Typetexp.Unbound_type_constructor type_longident))
      in
	    
      { exp_desc = Texp_typeinfo path;
	exp_loc = sexp.pexp_loc;
	exp_type = Predef.type_type_info;
	exp_env = env }
*)
  | Pexp_generic(rec_flag, spat_sexp_list, sbody) ->
      let (pat_exp_list, new_env) = 
	type_generic env Nonextensible rec_flag spat_sexp_list in
      let body = type_exp new_env sbody in
      { exp_desc = Texp_generic(rec_flag, new_env, pat_exp_list, body);
        exp_loc = sexp.pexp_loc;
        exp_type = body.exp_type;
        exp_env = env }
  | Pexp_coerce(sarg, caselist) ->
      let arg = type_expect env sarg Predef.type_dyn in
      let ty_res = newvar() in
      let cases = type_coerce_cases env ty_res caselist in
(* Usually it is partial !
      Parmatch.check_unused cases;
      Parmatch.check_partial sexp.pexp_loc cases;
*)
      { exp_desc = Texp_coerce(arg, cases);
        exp_loc = sexp.pexp_loc;
        exp_type = ty_res;
        exp_env = env }

(* Typing of an expression with an expected type.
   Some constructs are treated specially to provide better error messages. *)

and type_expect env sexp ty_expected =
  match sexp.pexp_desc with
    Pexp_constant(Const_string s as cst) ->
      let exp =
        { exp_desc = Texp_constant cst;
          exp_loc = sexp.pexp_loc;
          exp_type =
            (* Terrible hack for format strings *)
            begin match (repr ty_expected).desc with
              Tconstr(path, _, _) when Path.same path Predef.path_format ->
                type_format sexp.pexp_loc s
            | _ -> instance Predef.type_string
            end;
          exp_env = env } in
      unify_exp env exp ty_expected;
      exp
  | Pexp_let(rec_flag, spat_sexp_list, sbody) ->
      let (pat_exp_list, new_env) = type_let env rec_flag spat_sexp_list in
      let body = type_expect new_env sbody ty_expected in
      { exp_desc = Texp_let(rec_flag, pat_exp_list, body);
        exp_loc = sexp.pexp_loc;
        exp_type = body.exp_type;
        exp_env = env }
  | Pexp_sequence(sexp1, sexp2) ->
      let exp1 = type_statement env sexp1 in
      let exp2 = type_expect env sexp2 ty_expected in
      { exp_desc = Texp_sequence(exp1, exp2);
        exp_loc = sexp.pexp_loc;
        exp_type = exp2.exp_type;
        exp_env = env }
  | Pexp_function caselist ->
      let (ty_arg, ty_res) =
        try filter_arrow env ty_expected with Unify _ ->
          raise(Error(sexp.pexp_loc, Too_many_arguments))
      in
      let cases =
        List.map
          (fun (spat, sexp) ->
             let (pat, ext_env) = type_pattern env spat in
             unify_pat env pat ty_arg;
             let exp = type_expect ext_env sexp ty_res in
             (pat, exp))
          caselist
      in
      Parmatch.check_unused cases;
      Parmatch.check_partial sexp.pexp_loc cases;
      { exp_desc = Texp_function cases;
        exp_loc = sexp.pexp_loc;
        exp_type = newty (Tarrow(ty_arg, ty_res));
        exp_env = env }
  | _ ->
      let exp = type_exp env sexp in
      unify_exp env exp ty_expected;
      exp

(* Typing of statements (expressions whose values are discarded) *)

and type_statement env sexp =
    let exp = type_exp env sexp in
    match (expand_head env exp.exp_type).desc with
    | Tarrow(_, _) ->
        Location.print_warning sexp.pexp_loc Warnings.Partial_application;
        exp
    | Tconstr (p, _, _) when Path.same p Predef.path_unit -> exp
    | Tvar -> exp
    | _ ->
        Location.print_warning sexp.pexp_loc Warnings.Statement_type;
        exp

(* Typing of match cases *)

and type_cases env ty_arg ty_res caselist =
  List.map
    (fun (spat, sexp) ->
      let (pat, ext_env) = type_pattern env spat in
      unify_pat env pat ty_arg;
      let exp = type_expect ext_env sexp ty_res in
      (pat, exp))
    caselist

(* Typing of coerce match cases *)
and type_coerce_cases env ty_res caselist =
  List.map
    (fun (spat, sexp) ->
      Ctype.inside_coerce_typecase := true;
      begin_def ();
      let (pat, ext_env) = 
	try 
	  let result = type_pattern env spat in
	  Ctype.inside_coerce_typecase := false;
	  result
	with
	  e -> 
	    Ctype.inside_coerce_typecase := false;
	    raise e
      in
      end_def ();
      let typ = generalize pat.pat_type; pat.pat_type in
    Format.print_string "Pattern: "; 
    Printtyp.type_scheme typ;
    Format.print_newline ();
      let typ' = Ctype.instance typ in
      if not (moregeneral env true pat.pat_type typ') then
	raise (Error (pat.pat_loc, Generic_type_case_mismatch (pat.pat_type, typ)));
      let exp = type_expect ext_env sexp ty_res in
      (pat, typ, exp))
    caselist

(* Typing of let bindings *)

and type_let env rec_flag spat_sexp_list =
  begin_def();
  let (pat_list, new_env) =
    type_pattern_list env (List.map (fun (spat, sexp) -> spat) spat_sexp_list)
  in
  let exp_env =
    match rec_flag with Nonrecursive -> env | Recursive -> new_env in
  let exp_list =
    List.map2
      (fun (spat, sexp) pat -> 
	type_expect exp_env sexp pat.pat_type) spat_sexp_list pat_list in
  List.iter2
    (fun pat exp -> Parmatch.check_partial pat.pat_loc [pat, exp])
    pat_list exp_list;
  end_def();

  (* the patterns of the derived generic definitions must be "simple" *)
  let is_simple_pattern pat = 
    match pat.pat_desc with
    | Tpat_any _ -> true
    | Tpat_var (_,_) -> true
    | _ -> false
  in

  (* let x,y = generic_exp 
     cannot be generalized *)
  List.iter2
    (fun pat exp ->
       if not (is_nonexpansive exp) || 
          (Dtype.dynamic_variables exp.exp_type <> [] &&
	   not (is_simple_pattern pat))
       then
         iter_pattern (fun pat -> make_nongen pat.pat_type) pat)
    pat_list exp_list;
  
  (* We need to keep the original type of patterns. *) 
(*NEEDLESS?
  let original_pattern_variables = ref ([] : (Ident.t * type_expr) list) in
  let rec store_original_pat_type p =
    let enter_variable id ty =
      try
      	ignore (List.assoc id !pattern_variables);
      	raise (Failure "store_original_pat_type")
      with
      	Not_found ->
	  original_pattern_variables := (id, ty) :: !original_pattern_variables
    in
    match p.pat_desc with
      Tpat_any -> ()
    | Tpat_var id -> enter_variable id p.pat_type
    | Tpat_alias(p',id) ->
	store_original_pat_type p'; enter_variable id p.pat_type
    | Tpat_constant _ -> ()
    | Tpat_tuple pl -> List.iter store_original_pat_type pl
    | Tpat_construct (constr, args) -> List.iter store_original_pat_type args
    | Tpat_record largs -> List.iter (fun (_,arg) -> store_original_pat_type arg) largs
    | Tpat_array pl -> List.iter store_original_pat_type pl
    | Tpat_or (p1,p2) -> store_original_pat_type p1; store_original_pat_type p2
  in
  List.iter store_original_pat_type pat_list;
*)

  List.iter
    (fun pat -> iter_pattern (fun pat -> 
      pat.pat_type <- Dtype.generalize_copy pat.pat_type
	   (* generalize pat.pat_type *) ) pat)
    pat_list;

  (* generalized type variable table (org, generalized) *)
(*NEEDLESS?
  let varstbl = Dtype.generalized_variables () in
*)

  (* re-create new_env *)
  let pattern_variables = ref ([] : (Ident.t * type_expr) list) in
  let enter_variable id ty =
    try
      ignore (List.assoc id !pattern_variables);
      raise (Dtype.Generic_system_error "Something wrong in re-create new_env")
    with
      Not_found ->
	pattern_variables := (id, ty) :: !pattern_variables
  in
  let rec restore_pat p =
    match p.pat_desc with
      Tpat_any _ -> ()
    | Tpat_var (id,_) -> enter_variable id p.pat_type
    | Tpat_alias(p',id) ->
	restore_pat p'; enter_variable id p.pat_type
    | Tpat_constant _ -> ()
    | Tpat_tuple pl -> List.iter restore_pat pl
    | Tpat_construct (constr, args) -> List.iter restore_pat args
    | Tpat_record largs -> List.iter (fun (_,arg) -> restore_pat arg) largs
    | Tpat_array pl -> List.iter restore_pat pl
    | Tpat_or (p1,p2) -> restore_pat p1; restore_pat p2
  in
  (* modify new_env *)
  List.iter restore_pat pat_list;

  (* type schemes *)
  List.iter (fun (id, ty) ->
    let vdesc = Env.find_value (Path.Pident id) new_env in
    vdesc.val_type <- ty) !pattern_variables;

  List.combine pat_list exp_list, new_env

and type_generic env extensible_flag rec_flag bindings =
  (* begin_def();  USELESS (done in type_generic_pattern_list) *)
  let (pat_list, new_env) = 
    type_generic_pattern_list env extensible_flag bindings in
  let exp_env =
    match rec_flag with Nonrecursive -> env | Recursive -> new_env in
  let include_env = env in
  let caseset_list =
    List.map2 (fun (spat, sexp) pat ->
      let patpath = 
	match pat.pat_desc with
	| Tpat_var (id,_) -> Path.Pident id
	| _ -> assert false
      in
      type_generic_sub patpath exp_env include_env sexp pat.pat_type) 
      bindings pat_list 
  in

  (* exaustive check is not done here..., like let *)
  (* end_def(); USELESS (done in type_generic_pattern_list) *)
  (* they are all non-expansive... *)
  List.iter (fun pat -> iter_pattern (fun pat ->
          pat.pat_type <- Dtype.generalize_copy pat.pat_type
	   (* generalize pat.pat_type *) ) pat)
    pat_list;
  (* re-create new_env *)
  let pattern_variables = ref [] in
  let enter_variable id ty =
    try
      ignore (List.assoc id !pattern_variables);
      raise (Dtype.Generic_system_error "Something wrong in re-create new_env...")
    with
      Not_found ->
	pattern_variables := (id, ty) :: !pattern_variables
  in
  let rec restore_pat p =
    match p.pat_desc with
      Tpat_any _ -> ()
    | Tpat_var (id,_) -> enter_variable id p.pat_type
    | _ -> assert false
(*
    | Tpat_alias(p',id) ->
	restore_pat p; enter_variable id p.pat_type
    | Tpat_constant _ -> ()
    | Tpat_tuple pl -> List.iter restore_pat pl
    | Tpat_construct (constr, args) -> List.iter restore_pat args
    | Tpat_record largs -> List.iter (fun (_,arg) -> restore_pat arg) largs
    | Tpat_array pl -> List.iter restore_pat pl
    | Tpat_or (p1,p2) -> restore_pat p1; restore_pat p2
*)
  in
  (* modify new_env *)
  List.iter restore_pat pat_list;
  List.iter (fun (id, ty) ->
    let vdesc = Env.find_value (Path.Pident id) new_env in
    vdesc.val_type <- ty) !pattern_variables;

  List.combine pat_list caseset_list,
  new_env

and type_generic_sub path env include_env cases gscm =
  if debug then begin
    Format.print_string "Generic case typing";
    Format.print_newline ();
    Format.print_string "global type scheme (gscm) = ";
    Printtyp.type_scheme gscm;
    Format.print_newline ()
  end;

  (*
    generic (id : GSCM) = case
      PSCM => EXPR (type scheme = EXPR_SCM)
  *)

  let typ_cases = List.map (function
    | Pgeneric_typecase (sty,sexpr) ->
      (* check for each type case *)
	let pscm = Typetexp.transl_pattern_type_scheme env sty in

	if debug then begin
	  Format.print_string "  Type pattern (pscm) = "; 
	  Printtyp.type_scheme pscm;
	  Format.print_newline ();
	end;

	begin_def (); (* must be before pscm' creation for 
			 unifing expr.expr_typ pscm' *)
	let pscm', dvars_scm' = 
	  try
	    generic_moregeneral env gscm pscm 
	  with
	  |	_ ->
	      raise (Error (sty.ptyp_loc, 
			    Generic_type_case_mismatch (pscm, gscm)))
	in

	if debug then begin
	  Format.print_string "    Type checking expr (expr_scm)...\n";
	end;

        (* type the body *)
        (* suppose body is explicitly type constrained by pattern type *)
	let expr = 
	  let typ = Typetexp.transl_simple_type env false sty in
	  type_expect env sexpr typ
	in
	
	end_def ();

        (* generalize pscm and put Generalized None in dvars_scm' *)
	let pscm = Dtype.generalize_copy pscm' in
(*
    	if debug then begin
    	  Format.print_string "  pscm' generalization ";
    	  Printtyp.type_scheme pscm';
    	  Format.print_string " [ ";
    	  List.iter (fun typ -> 
    	    Printtyp.type_scheme typ;
    	    Format.print_string "; ") dvars_scm';
    	  Format.print_string " ]";
    	  Format.print_newline ()
    	end;
*)

        (* Unify expr.exp_type and type of pattern (pscm) 
           (we have to improve error message) *)
	begin try
	  unify env expr.exp_type pscm'
	with
	  _ -> raise (Error (expr.exp_loc, 
			     Generic_type_case_mismatch (expr.exp_type, pscm)))
	end;
    
        (* We generalize the type of expr *)
	if not (is_nonexpansive expr) then make_nongen expr.exp_type;
	let expr_scm = Dtype.generalize_copy expr.exp_type in

        (* expr_scm must be equal to pscm (we have to improve error message) *)
	if debug then begin
	  Format.print_string "    Check expr_scm = pscm : ";
	  Printtyp.type_scheme expr_scm;
	  Format.print_string " ?=? ";
	  Printtyp.type_scheme pscm;
	  Format.print_newline ()
	end;
	if not (Dtype.equal_scheme env expr_scm pscm) then begin
	  raise (Error (expr.exp_loc, 
			Generic_type_case_mismatch (expr_scm, pscm)))
	end;
        (* double check debug *)
	if debug then begin
	  Format.print_string "    Ok. Type case : ";
	  Printtyp.type_scheme pscm;
	  Format.print_string "    Dvars: ";
	  List.iter (fun dvar -> 
	    Printtyp.type_expr dvar; Format.print_string " ") dvars_scm';
	  Format.print_newline ();
	end;

	let varstbl = Dtype.generalized_variables () in

	Tgeneric_typecase { typecase_scheme= expr_scm;
			    typecase_inst= dvars_scm';
			    typecase_method= expr}
    | Pgeneric_include (lid,loc) ->
	(* we have done it in type_generic_pattern_list once *)
	(* lid must be searched in the original environment *)
	let (path,desc) = Env.lookup_value lid include_env in
	(* check gscm > desc.val_type *)

	begin try
	  ignore (generic_moregeneral env gscm desc.val_type)
	with
	| _ ->
	    raise (Error (loc,
			  Generic_type_case_mismatch (desc.val_type, gscm)))
	end;
	
	Tgeneric_include (path, desc, loc) ) cases	
      
  in
(*
  unify env scm pattyp; (* pattyp should be tvar so it is ok... *)
                        (* this does not affect scm *)
*)
  { gen_scheme= gscm;
    gen_cases= typ_cases }

(***************
(* This code must be unified with type_generic_sub *)
and type_generic_extension env orgtyp cases =
(*
Format.print_string "Original Type: ";
Printtyp.type_scheme orgtyp;
Format.print_newline ();
*)

  let type_cases = List.map (fun (sty,sexpr) ->
    let ty = Typetexp.transl_pattern_type_scheme env sty in
(*
Format.print_string "Pattern: "; 
Printtyp.type_scheme ty;
Format.print_string "  ==>  ";
*)
    (* typ > ty *)
    (* check typ > ty *)
    let typ' = instance orgtyp in
    let dvars = Dtype.dynamic_variables typ' in
    let ty' = instance ty in
    if not (moregeneral env true typ' ty') then begin
      raise (Error (sty.ptyp_loc, Generic_type_case_mismatch (ty, orgtyp)))
    end;
    Printtyp.type_expr ty';
    Format.print_newline ();

    (* I am really not sure it is correct... *)
    begin_def ();
    let expr = type_exp env sexpr in
    end_def ();
    
    if not (is_nonexpansive expr) then make_nongen expr.exp_type;
    let escm = Dtype.generalize_copy expr.exp_type in

(*
Format.print_string "Action: "; 
Printtyp.type_scheme escm;
Format.print_newline ();
*)

    if not (moregeneral env true escm ty') then begin
      raise (Error (expr.exp_loc, Generic_type_case_mismatch (escm, ty)))
    end;

    (* monomorphic check for create dvar link 
       --- this check must always succeed *)
    if not (moregeneral env true expr.exp_type ty') then begin
      raise (Error (expr.exp_loc, Generic_type_case_mismatch (escm, ty)))
    end;
    
(*
    Format.print_string "Dvars: ";
    List.iter (fun dvar -> Printtyp.type_expr dvar; Format.print_string " ") dvars;
    Format.print_newline ();
*)    
    (dvars, expr) ) cases
  in
  type_cases
*************************)


(* Typing of toplevel bindings *)

let type_binding env rec_flag spat_sexp_list =
  Typetexp.reset_type_variables();
  type_let env rec_flag spat_sexp_list

let type_generic_binding env rec_flag spat_sexp_list =
  Typetexp.reset_type_variables();
  type_generic env rec_flag spat_sexp_list

(* Typing of toplevel expressions *)

let type_expression env sexp =
  Typetexp.reset_type_variables();
  begin_def();
  let exp = type_exp env sexp in
  end_def();

  (* if is_nonexpansive exp then generalize exp.exp_type *)
  let gen_type = 
    if is_nonexpansive exp then 
      Dtype.generalize_copy exp.exp_type
    else begin make_nongen exp.exp_type; exp.exp_type end
  in

  exp, gen_type

(* Error report *)

open Format
open Printtyp

let report_error = function
    Unbound_value lid ->
      print_string "Unbound value "; longident lid
  | Unbound_constructor lid ->
      print_string "Unbound constructor "; longident lid
  | Unbound_label lid ->
      print_string "Unbound label "; longident lid
  | Constructor_arity_mismatch(lid, expected, provided) ->
      open_box 0;
      print_string "The constructor "; longident lid;
      print_space(); print_string "expects "; print_int expected;
      print_string " argument(s),"; print_space();
      print_string "but is here applied to "; print_int provided;
      print_string " argument(s)";
      close_box()
  | Label_mismatch(lid, trace) ->
      unification_error true trace
        (function () ->
           print_string "The label "; longident lid;
           print_space(); print_string "belongs to the type")
        (function () ->
           print_string "but is here mixed with labels of type")
  | Pattern_type_clash trace ->
      unification_error true trace
        (function () ->
           print_string "This pattern matches values of type")
        (function () ->
           print_string "but is here used to match values of type")
  | Multiply_bound_variable ->
      print_string "This variable is bound several times in this matching"
  | Orpat_not_closed ->
      print_string "A pattern with | must not bind variables"
  | Expr_type_clash trace ->
      unification_error true trace
        (function () ->
           print_string "This expression has type")
        (function () ->
           print_string "but is here used with type")
  | Apply_non_function typ ->
      begin match (repr typ).desc with
        Tarrow(_, _) ->
          print_string "This function is applied to too many arguments"
      | _ ->
          print_string
            "This expression is not a function, it cannot be applied"
      end
  | Label_multiply_defined lid ->
      print_string "The label "; longident lid;
      print_string " is defined several times"
  | Label_missing ->
      print_string "Some labels are undefined"
  | Label_not_mutable lid ->
      print_string "The label "; longident lid;
      print_string " is not mutable"
  | Bad_format s ->
      print_string "Bad format `"; print_string s; print_string "'"
  | Undefined_method (ty, me) ->
      reset (); mark_loops ty;
      open_vbox 0;
      open_box 0;
      print_string "This expression has type";
      print_break 1 2;
      type_expr ty;
      close_box ();
      print_cut ();
      print_string "It has no method ";
      print_string me;
      close_box ()
  | Undefined_inherited_method me ->
      print_string "This expression has no method ";
      print_string me
  | Unbound_class cl ->
      print_string "Unbound class "; longident cl
  | Virtual_class cl ->
      print_string "One cannot create instances of the virtual class ";
      longident cl
  | Unbound_instance_variable v ->
      print_string "Unbound instance variable ";
      print_string v
  | Instance_variable_not_mutable v ->
      print_string " The instance variable "; print_string v;
      print_string " is not mutable"
  | Not_subtype(tr1, tr2) ->
      reset ();
      List.iter
        (function (t, t') -> mark_loops t; if t != t' then mark_loops t')
        tr1;
      List.iter
        (function (t, t') -> mark_loops t; if t != t' then mark_loops t')
        tr2;
      trace true (fun _ -> print_string "is not a subtype of type") tr1;
      trace false (fun _ -> print_string "is not compatible with type") tr2
  | Outside_class ->
      print_string "This object duplication occurs outside a method definition."
  | Value_multiply_overridden v ->
      print_string "The instance variable "; print_string v;
      print_string " is overridden several times"
  | Coercion_failure (ty, ty', trace) ->
      unification_error true trace
        (function () ->
           mark_loops ty; if ty' != ty then mark_loops ty';
           print_string "This expression cannot be coerced to type";
           print_break 1 2;
           type_expansion ty ty';
           print_string ";";
           print_space ();
           print_string "it has type")
        (function () ->
           print_string "but is here used with type")
  | Too_many_arguments ->
      print_string "This function has too many arguments"
  | Scoping_let_module(id, ty) ->
      reset (); mark_loops ty;
      print_string "This `let module' expression has type";
      print_space(); type_expr ty; print_space();
      print_string "In this type, the locally bound module name ";
      print_string id; print_string " escapes its scope"
  | Masked_instance_variable lid ->
      print_string "The instance variable "; longident lid; print_space ();
      print_string
        "cannot be accessed from the definition of another instance variable"

  | Generic_not_closed ty ->
      print_string "This generic value has a type scheme";
      print_space ();
      type_scheme ty;
      print_space ();
      print_string "which is not generalizable";
  | Generic_illegal_approximation (t1,t2) ->
      print_string "The explicit generic approximation";
      print_space ();
      type_scheme t1;
      print_space ();
      print_string "must be more general than";
      print_space ();
      type_scheme t2
  | Generic_illegal_pattern ->
      print_string "A generic pattern must be a variable."
  | Generic_not_generic ty ->
      print_string "This has a type";
      print_space(); type_expr ty;
      print_string ", which is not polymorphic at all"
  | Generic_type_case_mismatch (t1, t2) ->
      print_string "This has a type";
      print_space(); type_scheme t1;
      print_string ", which does not match with the pattern type";
      print_space(); type_scheme t2
  | Generic_type_case_not_generalizable t1 ->
      print_string "This has a type";
      print_space(); type_scheme t1;
      print_string ", which is not fully generalizable"
  | Coerce_pattern_needs_type_info (t,v) ->
      print_string "This coerce pattern typed";
      print_space(); type_expr t; print_space();
      print_string "needs an explicit type constraint for a variable";
      print_space(); type_expr v
(*
  | No_match_konstraint (p,typ) ->
      print_string "No possible case matches with";
      print_space ();
      path p;
      print_string " : ";
      type_expr typ
  | Ambiguous_konstraints ptyps ->
      print_string "More of one possible constraints for";
      print_space ();
      List.iter (fun (p,typ) ->
	path p;
      	print_string " : ";
      	type_expr typ;
	print_space ()) ptyps
*)
