open Misc
open Longident
open Path
open Parsetree
open Types
open Typedtree
open Asttypes

open Format

open Dtype

let debug = try ignore (Sys.getenv "DEBUG_TYPEGENERIC"); true with _ -> false
let debug2 = try ignore (Sys.getenv "DEBUG_TYPEGENERICLOOP"); true with _ -> false

let new_par () =
  force_newline ();
(*
  print_string "  ";
*)
  print_string "  ";
  open_box 0
;;

let end_par () =
  close_box ();
  force_newline ();
;;

let rec print_flow f = 
  match f with
  | Flow_node node ->
      printf "#%d(id=%d)" node.flow_pos node.flow_id;
      if node.flow_sub <> [] then begin 
	printf " -";
	open_box 0;
	let rec printer = function 
	  | [] -> ()
	  | [f] -> print_string "\- "; print_flow f
	  | x::xs -> 
	      print_string "+- "; 
	      print_flow x; 
	      force_newline (); 
	      printer xs
	in
	printer node.flow_sub;
	close_box ()
      end
  | Flow_loop id -> printf "loop=>%d" id
  | Flow_type typs -> 
      let first = ref true in
      print_string "<<"; 
      List.iter (fun typ -> 
	if not !first then print_string ", ";
	first := true;
	Printtyp.type_expr typ) typs; 
      print_string ">>"
;;

(******************* extraction of generic related definitions and instances *)
type gabstract =
  | GenInst of expression  
  | GenLet of Env.t * Ident.t list * pattern list * expression * gabstract list
  | GenGeneric of Env.t * Ident.t list * pattern * 
	(generic_case * gabstract list) list
  | GenEval of expression * type_expr * 
	         generic_konstraints ref * generic_info * gabstract list
;;

(*************************************************************** misc. tools *)
let patterns pattern =
  let vars = ref [] in
  let rec sub pattern =
    match pattern.pat_desc with
    | Tpat_any _ 
    | Tpat_var (_, _)  -> vars := pattern :: !vars
    | Tpat_alias (p, id) -> sub p; vars := pattern :: !vars
    | Tpat_tuple pats -> List.iter sub pats
    | Tpat_construct (_,pats) -> List.iter sub pats
    | Tpat_record lpats -> List.iter (fun (_,p) -> sub p) lpats
    | Tpat_array pats -> List.iter sub pats
    | Tpat_or (p1,p2) -> sub p1; sub p2
    | _ -> assert false
  in
  sub pattern;
  !vars
;;

let pattern_idents pattern =
  let vars = ref [] in
  let rec sub pattern =
    match pattern.pat_desc with
    | Tpat_any _ -> () 
    | Tpat_var (id, _)  -> vars := id :: !vars
    | Tpat_alias (p, id) -> sub p; vars := id :: !vars
    | Tpat_tuple pats -> List.iter sub pats
    | Tpat_construct (_,pats) -> List.iter sub pats
    | Tpat_record lpats -> List.iter (fun (_,p) -> sub p) lpats
    | Tpat_array pats -> List.iter sub pats
    | Tpat_or (p1,p2) -> sub p1; sub p2
    | _ -> assert false
  in
  sub pattern;
  !vars
;;

let get_ident exp =
  match exp.exp_desc with
  | Texp_ident (p, vdesc,_) -> p, vdesc
  | _ -> assert false
;;

let path_of gid = fst (get_ident gid)

(*************************************** konstraint coping and instantiation *)
let copy_konstraint cache konst =
  let cas = copy_type cache konst.konst_case in
  let derived = 
    List.map (fun kd -> { kd with derived_type= copy_type cache kd.derived_type}) konst.konst_derived
  in
  { konst_case= cas;
    konst_derived= derived
  } 
;;

let instance_konstraint konst =
  let typs = List.map Ctype.copy (konst.konst_case :: 
				  (List.map (fun dk -> dk.derived_type)
				     konst.konst_derived)) in
  Btype.cleanup_types ();
  List.iter Btype.unmark_type typs;
  match typs with
  | c :: ders ->
      { konst_case= c;
	konst_derived= 
	  List.map2 (fun kd t -> 
	    {kd with derived_type = t}) konst.konst_derived ders }
  | _ -> assert false
;;

(******************************* Unification test (not actually unify types) *)

let unifiable env typ1 typ2 =
  (* Not really unify them, just look they are unifiable or not *)
  match copy_types Reset [typ1; typ2] with (* tdvar links are reset *)
  | [typ1'; typ2'] ->
      begin try Ctype.unify env typ1' typ2'; true 
      with Ctype.Unify _ -> false end
  | _ -> assert false
;;

let generic_moregeneral env typ1 typ2 =
  match copy_types Reset [typ1; typ2] with
  | [typ1';typ2'] -> 
      let typ1' = generalize_copy typ1'
      and typ2' = generalize_copy typ1'
      in
      Ctype.moregeneral env false typ1' typ2' 
  | _ -> assert false
;;

(******************************* extract generic abstraction from expression *)
let rec find_generics_exp e =
  match e.exp_desc with
  | Texp_let (recflag,pexps,exp) ->
      let recursives =
	if recflag = Recursive then 
	  List.flatten (List.map (fun (p,e) -> pattern_idents p) pexps)
	else []
      in
      List.flatten (List.map (fun (p,e) -> 
	if Dtype.generalized_dynamic_variables p.pat_type <> [] 
	then begin
	  [GenLet(exp.exp_env, recursives, patterns p, e, find_generics_exp e)]
	end else find_generics_exp e) pexps) @
      find_generics_exp exp
  | Texp_generic (recflag, env', bindings, exp) ->
      let recursives =
	if recflag = Recursive then 
	  List.flatten (List.map (fun (p,_) -> pattern_idents p) bindings)
	else []
      in
      (List.map (fun (p,caseset) ->
	let pat = 
	  match patterns p with
	  | [pat] -> pat
	  | _ -> assert false
	in
	GenGeneric (env' (* should be equal to exp.exp_env *),
		    recursives,
		    pat,
		    List.map (fun c ->
		      match c with
		      | Tgeneric_typecase cas -> 
			  c, find_generics_exp cas.typecase_method
		      | Tgeneric_include _ -> 
			  c, []) caseset.gen_cases)) bindings) @
      find_generics_exp exp
  | Texp_ident (path,vdesc,_) ->
      if Dtype.generalized_dynamic_variables vdesc.val_type <> []
      then [GenInst e] else []
  | Texp_constant _ -> []
  | Texp_function pexps ->
      List.flatten (List.map (fun (_,e) -> find_generics_exp e) pexps)
  | Texp_apply (e,exps) ->
      List.flatten (List.map find_generics_exp (e::exps))
  | Texp_match (e, pexps) ->
      find_generics_exp e @ 
      (List.flatten (List.map (fun (_,e) -> find_generics_exp e) pexps))
  | Texp_try (e, pexps) ->
      find_generics_exp e @ 
      (List.flatten (List.map (fun (_,e) -> find_generics_exp e) pexps))
  | Texp_tuple exps -> List.flatten (List.map find_generics_exp exps)
  | Texp_construct (_, exps) -> List.flatten (List.map find_generics_exp exps)
  | Texp_record (lexps, eopt) ->
      List.flatten (List.map (fun (_,e) -> find_generics_exp e) lexps) @ 
      begin match eopt with
	Some e -> find_generics_exp e
      | None -> []
      end
  | Texp_field (e,_) -> find_generics_exp e
  | Texp_setfield (e1, _, e2) -> find_generics_exp e1 @ find_generics_exp e2
  | Texp_array exps -> List.flatten (List.map find_generics_exp exps)
  | Texp_ifthenelse (e1,e2,eopt) ->
      find_generics_exp e1 @ find_generics_exp e2 @
      begin match eopt with
	Some e -> find_generics_exp e
      | None -> []
      end
  | Texp_sequence (e1,e2) -> find_generics_exp e1 @ find_generics_exp e2
  | Texp_while (e1, e2) -> find_generics_exp e1 @ find_generics_exp e2
  | Texp_for (_, e1, e2, _, e3) ->
      find_generics_exp e1 @ find_generics_exp e2 @ find_generics_exp e3
  | Texp_when (e1,e2) -> find_generics_exp e1 @ find_generics_exp e2
  | Texp_override (_, pexps) ->
      List.flatten (List.map (fun (_,e) -> find_generics_exp e) pexps)
  | Texp_letmodule (_, (* module_expr is done *) _, e) -> find_generics_exp e
  | Texp_coerce (e, ptexps) ->
      find_generics_exp e @ 
      (List.flatten (List.map (fun (_,_,e) -> find_generics_exp e) ptexps))
  | _ -> assert false
  (*
  | Texp_send (e,_) -> find_generics_exp e
  | Texp_new _ -> []
  | Texp_instvar _ -> []
  | Texp_setinstvar (_,_,e) -> find_generics_exp e
  *)
;;

let find_generics_str env strs =
  let find_str str =
    match str with
    | Tstr_eval (e,scm,kref,ginfo) -> 
	if Dtype.generalized_dynamic_variables scm <> [] then
	  [GenEval (e, scm, kref, ginfo, find_generics_exp e)]
	else find_generics_exp e
    | Tstr_value (recflag, pexps) -> 
	let recursives =
	  if recflag = Recursive then 
	    List.flatten (List.map (fun (p,_) -> pattern_idents p) pexps)
	  else []
	in
	List.flatten (List.map (fun (p,e) ->
	  if Dtype.generalized_dynamic_variables p.pat_type <> [] 
	  then [GenLet(env, recursives, patterns p, e, find_generics_exp e)]
	  else find_generics_exp e) pexps)
    | Tstr_generic (_,recflag,env',bindings) ->
	let recursives =
	  if recflag = Recursive then 
	    List.flatten (List.map (fun (p,_) -> pattern_idents p) bindings)
	  else []
	in
	List.map (fun (p,caseset) ->
	  let pat =
	    match patterns p with
	    | [pat] -> pat
	    | _ -> assert false
	  in
	  GenGeneric (env (* > env' *), recursives, pat,
		      (List.map (fun c ->
			match c with
			| Tgeneric_typecase cas -> 
			    c, find_generics_exp cas.typecase_method
			| Tgeneric_include _ ->
			    c, [] ) 
			 caseset.gen_cases))) bindings
    | Tstr_genextend (_,_,_,cases) -> assert false (* not yet *)
 (* | Tstr_module is done of its typing ? *)
    | _ -> [] (* assert false *) (* I do not think about classes... *)
  in
  List.flatten (List.map find_str strs)
;;

(****************************************************** konstraint fixiation *)
(* make konstraints well formed *)

(* filter polymorphic instances related to gendvars *)

let filter_pinsts gendvars pinsts = 
  let rec contents dvars =
    try
      List.iter (fun gendvar ->
	match gendvar.desc with
	| Tdvar (info,_) ->
	    List.iter (fun dvar ->
	      match dvar.desc with
	      | Tdvar (info',_) ->
		  if info == info' then raise Exit
	      | _ -> assert false) dvars
	| _ -> assert false) gendvars;
      false
    with
    | Exit -> true
  in

  List.fold_right (fun pinst (related,notrelated) ->
    let p, vdesc = get_ident pinst in
    let dv1 = Dtype.dynamic_variables pinst.exp_type in
    let dv2 = Dtype.dynamic_variables vdesc.val_type in
    if contents dv1 || contents dv2 then
      (pinst :: related, notrelated)
    else
      (related, pinst :: notrelated)) pinsts ([],[])
;;
      
(* fix recursive occurrences of konstraints to avoid inf. loop
   of signature export *)

let rec replace_rec_konst_derived recset dk = 
  match dk.derived_desc with
  | None -> dk
  | Some desc -> 
      if List.mem dk.derived_path recset then {dk with derived_desc= None}
      else dk

and replace_rec_konstraints recset = function
  | Konst_None -> Konst_None
  | Konst_Cases (p,cks) ->
      Konst_Cases (p,
        List.map (fun ck ->
	  { ck with konst_derived= 
                 List.map (replace_rec_konst_derived (p::recset)) 
                  ck.konst_derived }) cks)
  | Konst_Derived (p,dks) ->
      Konst_Derived (p, 
        List.map (replace_rec_konst_derived (p::recset)) dks) 
;;

(* polyinsts dvars are different from their corresponding generalized
   dvars. they must be same in the konstraints *)
let fix_derived konst =
  let gendvars = Dtype.generalized_dynamic_variables konst.konst_case in
  let info_dvars = List.map (fun dvar -> 
    match dvar.desc with
    | Tdvar (x,_) -> x, dvar
    | _ -> assert false) gendvars 
  in
  let rec sub ty = 
    match ty.desc with
    | Tdvar (x,_) ->
	begin try
	  List.assq x info_dvars 
	with
	  Not_found -> ty
	end
    | Tarrow (t1,t2) -> { ty with desc = Tarrow( sub t1, sub t2 ) }
    | Ttuple tys -> { ty with desc = Ttuple( List.map sub tys ) }
    | Tconstr (p, tys, abref) ->
	 { ty with desc = Tconstr (p, List.map sub tys, abref) }
    | Tlink ty -> sub ty
    | Tvar -> ty
    | _ -> assert false
  in

  let levfix ty =
(* NOTE !!!
   If a type contains generalized type variable inside,
   then its level must be GENERIC_LEVEL !!! 
*)
    let rec levfix_aux result ty =
      let have_gen = ref false in
      Btype.iter_type_expr (levfix_aux have_gen) ty;
      if !have_gen then ty.level <- Btype.generic_level;
      if ty.level = Btype.generic_level then result := true
    in
    levfix_aux (ref false) ty;
    ty
  in
  { konst with 
      konst_derived = List.map (fun kd -> 
	{ kd with derived_type=  levfix (sub kd.derived_type) })
          konst.konst_derived}
;;

(**************************************************** konstraint assignments *)

exception Not_admissible of expression * type_expr
exception Not_defined_by_generic of Path.t * Location.t

let debug_assign id typ k = 
  Printtyp.ident id;
  print_string " ";
  Printtyp.extended_scheme Env.empty typ k;
  force_newline ()

let set_compilation_layout konsts ginfo =
  match konsts with
  | Konst_Cases (path, cks) ->
      ginfo.ginfo_layout <- Some (Ktype.generic_compilation_layout konsts)
  | _ -> ()
;;

let rec assign gabsts = 
  let rec sub = function
    | GenInst exp -> [exp]
    | GenLet (env, recursives, patterns, exp, gabsts) ->
	assign_let env recursives patterns exp gabsts
    | GenEval ({exp_desc= Texp_ident (p,vdesc,iref)}, 
	       scm, kref, ginfo, gabsts) ->
	(* the ident must be a polymorphic instance *)
	iref := Inst_poly_generic { poly_parent= ginfo; (*?*)
				    poly_pos= 0;
				    poly_call = None };
	(* for queries of generic konstraints *)
	kref := vdesc.val_konstraints;
	if debug then debug_assign (Ident.create "-") 
	                 vdesc.val_type !kref;
	[]
    | GenEval (exp, scm, kref, ginfo, gabsts) ->
	assign_eval_nonident exp scm kref ginfo gabsts
    | GenGeneric (env, recursives, pat, cases) ->
	assign_generic env recursives pat cases 

  in
  List.flatten (List.map sub gabsts)

and assign_let env recursives patterns exp gabsts =
  (* global filtering *)
  let gendvars = 
    List.flatten (List.map (fun pat ->
      Dtype.generalized_dynamic_variables pat.pat_type) patterns)
  in
  let related_all,notrelated = filter_pinsts gendvars (assign gabsts) in
  (* for each pattern variable *)

  List.iter (fun pat ->
    let idopt, ginfo = 
      match pat.pat_desc with
      | Tpat_any ginfo -> None, ginfo
      | Tpat_var (id,ginfo) -> Some id, ginfo
      | _ -> 
	  prerr_endline "GenLet?";
	  assert false (* possible bug *)
    in
    let gendvars_id = Dtype.generalized_dynamic_variables pat.pat_type in
    let related_to_id,_ = filter_pinsts gendvars_id related_all in
	  
    (* assign Inst_poly_generic id *)
    let pos = ref 0 in
    List.iter (fun exp -> 
      match exp.exp_desc with
      | Texp_ident (p,_,iref) ->
	  begin match !iref with
	  | Inst_normal -> 
	      iref := Inst_poly_generic { poly_parent= ginfo;
					  poly_pos = !pos;
					  poly_call = None }
	  | Inst_poly_generic _ ->
	      raise (Dtype.Generic_system_error "A poly inst depends on more than one patvars!")
	  | Inst_full_generic _ -> assert false
	  end;
	  incr pos
      | _ -> assert false ) related_to_id;
    
    let derived = List.map (fun e ->
      let path, vdesc = get_ident e in
      if debug then begin
	try 
	  ignore (Env.find_value path env);
	with
	| Not_found ->
	    prerr_endline ("Warning: the constraint of " ^ Path.string_of path ^ " is trying to escape from its scope (this is ok, but not well tested...)");
      end;
      { derived_path= path;
	derived_desc= Some vdesc;
	derived_type= e.exp_type }) related_to_id
    in

    let sizes = List.fold_right (fun e st ->
      let path, vdesc = get_ident e in
      match path with
      | Pident id -> 
	  if List.mem id recursives then
	    (e, Dtype.size_of_instance_predicate 
	       env e.exp_type vdesc.val_type)
	    :: st
	  else st
      | _ -> st) related_to_id []
    in

    match idopt with
    | Some id ->
	if List.mem id recursives then begin
	  let size = 
	    Dtype.size_of_instance_predicate env 
	      pat.pat_type pat.pat_type 
	  in
	  List.iter (fun (e,size') ->
	    if not (Dtype.leq size' size) then 
	      raise (Not_admissible (e,pat.pat_type) ))
	    sizes;
	end;

	let vdesc = Env.find_value (Pident id) env in
	let konst = 
          (* badly written *)
	  Konst_Derived (Pident id,
		 (fix_derived { konst_case= pat.pat_type; 
				konst_derived= derived }).konst_derived)
	in
	vdesc.val_konstraints <- replace_rec_konstraints [] konst;
	if debug then 
	  debug_assign id vdesc.val_type vdesc.val_konstraints
    | None -> () ) patterns;

  notrelated

and assign_eval_nonident exp scm kref ginfo gabsts =

  let id = Ident.create "-" in
  let gendvars = Dtype.generalized_dynamic_variables scm in
  let related,notrelated = filter_pinsts gendvars (assign gabsts) in

  let pos = ref 0 in
  List.iter (fun exp -> 
    match exp.exp_desc with
    | Texp_ident (p,_,iref) ->
	begin match !iref with
	| Inst_normal -> 
	    iref := Inst_poly_generic { poly_parent= ginfo;
					poly_pos= !pos;
					poly_call= None }
	| Inst_poly_generic _ ->
	    raise (Dtype.Generic_system_error "A poly inst depends on more than one patvars!")
	| Inst_full_generic _ -> assert false
	end;
	incr pos
    | _ -> assert false ) related;
  
  let derived = List.map (fun e ->
    let path,vdesc = get_ident e in
    { derived_path= path;
      derived_desc= Some vdesc;
      derived_type= e.exp_type } ) related
  in

  let konst = 
    Konst_Derived ( Pident id,
	            (* badly written *)
		   (fix_derived
		      { konst_case= scm;
			konst_derived= derived }).konst_derived)
  in
  let konst = replace_rec_konstraints [] konst in
  kref := konst;
  if debug then debug_assign id scm !kref;
  
  notrelated

and assign_generic env recursives pat cases =
    let idopt, ginfo = 
      match pat.pat_desc with
      | Tpat_any ginfo -> None, ginfo
      | Tpat_var (id,ginfo) -> Some id, ginfo
      | _ -> 
	  prerr_endline "GenGeneric?";
	  assert false (* possible bug *)
    in

    let notrelated_list = ref [] in
    let konstraint =
      List.flatten (List.map (function (cas, gabsts) ->
	match cas with
	| Tgeneric_typecase cas ->
	    let gendvars = Dtype.generalized_dynamic_variables 
		cas.typecase_scheme
	    in
	    let related,notrelated = 
	      filter_pinsts gendvars (assign gabsts) 
	    in

	    (* assign Inst_poly_generic id *)
	    let pos = ref 0 in
	    List.iter (fun exp -> 
	      match exp.exp_desc with
	      | Texp_ident (p,_,iref) ->
		  begin match !iref with
		  | Inst_normal -> 
		      iref := Inst_poly_generic { poly_parent= ginfo;
						  poly_pos= !pos;
						  poly_call= None }
		  | Inst_poly_generic _ ->
		      raise (Dtype.Generic_system_error 
			       "A poly inst depends on more than one patvars!")
		  | Inst_full_generic _ -> assert false
		  end;
		  incr pos
	      | _ -> assert false ) related;

	    notrelated_list := notrelated @ !notrelated_list;
	    let derived = List.map (fun e ->
	      let path,vdesc = get_ident e in
	      { derived_path= path;
		derived_desc= Some vdesc;
		derived_type= e.exp_type }) related
	    in

	    let sizes = List.fold_right (fun e st ->
	      let path, vdesc = get_ident e in
	      match path with
	      | Pident id -> 
		  if List.mem id recursives then
	  	    (e, Dtype.size_of_instance_predicate 
		       env e.exp_type vdesc.val_type)
		    :: st
		  else st
	      | _ -> st) related []
	    in

	    begin match idopt with
	    | Some id ->
		if List.mem id recursives then begin
		  let size = 
		    Dtype.size_of_instance_predicate env 
		      cas.typecase_scheme pat.pat_type 
		  in
		  List.iter (fun (e,size') ->
	      	    if not (Dtype.leq size' size) then 
		      raise (Not_admissible (e,cas.typecase_scheme) ))
	      	    sizes;
		end
	    | None -> ()
	    end;
	    
	    [ fix_derived
		{ konst_case = cas.typecase_scheme;
		  konst_derived= derived } ]
	| Tgeneric_include (path,desc,loc) ->
	    (* it must be defined by a generic binding *)
	    let cks =
	      match desc.val_konstraints with
	      | Konst_Cases (_,  cks) -> cks 
	      | _ -> raise (Not_defined_by_generic (path,loc))
	    in
            (* we scan cks and replace recursive occurrence of path
	       to our path *)
	    let mypath = 
	      match idopt with
	      | Some id -> Pident id
	      | None -> raise (Generic_system_error "include with nameless generics?")
	    in
	    let rec replace_path_case ck = 
	      { konst_case = ck.konst_case;
		konst_derived= List.map replace_path_derived 
		  ck.konst_derived }
	    and replace_path_derived dk =
	      let path' = 
		if dk.derived_path = path then mypath else dk.derived_path
	      in
	      { derived_path= path';
		derived_type= dk.derived_type;
		derived_desc= 
		begin
		  match dk.derived_desc with
		  | None -> None
		  | Some vdesc when mypath = path' -> 
 	              (* avoid inf loop *)
		      None
		  | Some vdesc (* when mypath <> path' *) -> 
		      Some (replace_path_vdesc vdesc)
		end }
	    and replace_path_vdesc vdesc =
	      { val_type = vdesc.val_type;
		val_kind = vdesc.val_kind;
		val_konstraints = 
		replace_path_konstraints vdesc.val_konstraints
	      } 
	    and replace_path_konstraints = function
	      | Konst_Cases (p,cks) ->
		  let p' = if p = path then mypath else p
		  in
		  Konst_Cases (p', List.map replace_path_case cks)
	      | Konst_Derived (p,dks) ->
		  let p' = if p = path then mypath else p
		  in
		  Konst_Derived (p', List.map replace_path_derived dks)
	      | Konst_None -> Konst_None
	    in
	    List.map replace_path_case cks) cases)
    in

    begin match idopt with
    | Some id ->
	let vdesc = Env.find_value (Pident id) env in
	let konst = 
	  replace_rec_konstraints [] 
	    (Konst_Cases (Pident id, konstraint))
	in
	vdesc.val_konstraints <- konst;
	if debug then debug_assign id vdesc.val_type vdesc.val_konstraints;
    | None -> ()
    end;

    !notrelated_list
;;

(************************************************************ reconstruction *)
exception No_match
exception No_match_with of Path.t * value_description * type_expr
exception Invalid_full_instance of 
  expression * Path.t * value_description * type_expr
exception Fixed of type_expr * generic_konstraint * int
exception Polyrec_restriction
exception Generic_expr_type_clash of Location.t * (type_expr * type_expr) list

type log = { paths: Path.t list; 
	     type_grown: Path.t list;
	     cache: ((Path.t * type_expr) * int) list }

let empty_log = { paths= []; type_grown= []; cache= [] }

let node_id = ref 0

let mode_choose_first_match = ref false

(* Check t <: \sigma d *)
(* returns the inferred type, ambiguity, and flow *)
let rec reconst_konst_case log env recvdescs layout path typ konst =
  let simple_debug = konst.konst_derived = [] in
  if debug then begin
    if not simple_debug then begin
      print_string "?? ";
(*
      Printtyp.path path; 
      print_string " : ";
*)
      Printtyp.type_expr typ;
      print_string " <: ";
      Printtyp.type_scheme konst.konst_case;
(*
      print_string " <: ";
      Printtyp.konstraints [konst];
*)
      new_par ();
    end
  end;

  (* init: copy types *)
  let copy_cache = create_copy_cache Duplicate in
  
  let orgtyp = typ in
  let orgkonst = konst in

  let typ = copy_type copy_cache typ in
  let log = { log with cache = List.map (fun ((p,t),id) -> 
                          (p, copy_type copy_cache t),id) log.cache} in
  let konst = copy_konstraint copy_cache konst in

  let konst_before_instance = konst in
  let konst = instance_konstraint konst in

  (* konst is instansiated *)

  let dvars = dynamic_variables typ in

  try
    begin try
      (* typ and konst are copied: unification is not destructive *)
      Ctype.unify env typ konst.konst_case; (* may fail *)
    with e ->
      if debug && (not simple_debug) then print_string "Not unifiable";
      raise e
    end;

    let type_grown =
      try List.iter (fun dv ->
	let s = Dtype.size dv in 
	let grown = 
	  match List.length s.tdvars with
	  | 0 -> false
	  | 1 -> s.const <> 0
	  | _ -> true
	in
	if grown then raise Exit) dvars;
  	false
      with
      | Exit -> true
    in

    if type_grown && debug then begin
      print_string "TYPE SIZE GROWN";
      force_newline ();
    end;

    let flows = 
      if type_grown (* && List.mem path log.type_grown *) then begin
	if debug then begin
	  print_string "POLYREC RESTRICTION !!!";
	  force_newline ();
        end;
        None
      end else begin
(*
	let log = 
          if type_grown = {log with type_grown= path :: log.type_grown} 
	  else log
        in
*)

	if debug && (not simple_debug) then begin
	  print_string "!! Unifiable  "; 
	  if orgkonst.konst_derived <> [] then begin
	    force_newline ();
	    print_string "Derivative check";
	    force_newline ();
	    Printtyp.type_expr typ;
	    print_string " ?=? ";
	    Printtyp.konstraints env 
	      (Konst_Cases (Pident (Ident.create "dummy"),
			    [konst]));
	    force_newline ();
	  end;
	end;

	let flows_rev = 
	  if debug && orgkonst.konst_derived <> [] then new_par ();
	  let end_debug () = 
	    if debug && orgkonst.konst_derived <> [] then end_par ()
	  in
	  try
	    let flows_rev, unifpairs =
	      List.fold_left (fun (frev,upairs) {derived_path= path; 
					derived_desc= vdesc; 
					derived_type= t}  -> 
		  let vdesc =
		    match vdesc with
		    | None -> List.assoc path recvdescs
		    | Some vdesc -> vdesc
		  in
		  let layout' =
		    if List.mem_assoc path layout then layout
		    else Ktype.generic_compilation_layout vdesc.val_konstraints
		  in
		  let t',_,flow = (* may fail *) 
		    reconst_path log env recvdescs layout' t path vdesc 
		  in 
		  if debug then begin print_string "&&"; force_newline () end;
		  (flow :: frev), ((t,t')::upairs)) ([],[]) konst.konst_derived
	    in
	    List.iter (fun (t,t') -> Ctype.unify env t t') unifpairs;
	    end_debug ();
	    flows_rev
	  with
	  | e -> end_debug (); raise e
	in
 
	try
	  Some (List.rev (List.map (function 
	    | Some f -> f 
	    | None -> raise Exit) flows_rev))
	with
	| Exit -> None
      end
    in

    if debug then begin
      if not simple_debug then end_par ();

      printf "%s: " (if flows = None then "?" else "O");
      Printtyp.type_expr typ;
      print_string " <: ";
      Printtyp.type_scheme orgkonst.konst_case;
(*
      Printtyp.konstraints [konst];
*)
      force_newline ();
      force_newline ();
    end;

    (* returns the inferred type, ambiguity, and flow *)
    typ, 
    not (!mode_choose_first_match && flows <> None),
    flows 
  with
  | e -> 
      if debug && (not simple_debug) then end_par ();
      if debug then begin
	printf "X: ";
	Printtyp.type_expr orgtyp;
	print_string " <: ";
	Printtyp.type_scheme orgkonst.konst_case;
(*
	Printtyp.konstraints [orgkonst];
*)
	force_newline ();
      end;
      raise e

(* Check t <: [..|..|..] *)
(* returns the inferred type and flow *)
and reconst_konst_cases log env recvdescs layout path typ konsts =

  (* cached ? *)
  let copy_cache = create_copy_cache Duplicate in 

  let orgtyp = typ in
  let orgkonsts = konsts in

  let typ = copy_type copy_cache typ in
  let log = { log with cache= List.map (fun ((p,t),id) ->
                  (p, copy_type copy_cache t),id) log.cache } in
  let konsts = List.map (copy_konstraint copy_cache) konsts in

  let found = ref [] in 
  
  begin try
    let pos = ref (-1) in
    List.iter (fun k ->
      incr pos;
      begin try
  	let t,is_ambiguous,flows = 
	  reconst_konst_case log env recvdescs layout path typ k in
  	found := (!pos,t,flows) :: !found;
	if not is_ambiguous then begin
	  if !mode_choose_first_match then begin
	    if List.length!found = 1 then begin
	      (* and if it is the first match *)
	      if debug then begin
		print_string "THIS CASE MUST BE SELECTED";
		force_newline ();
	      end;
  	      found := [!pos,t,flows]; (* cancel the other matches ! *)
	      raise Exit; (* The first match is fixed *)
	    end
	  end else begin
	    if debug then begin
	      print_string "THE OTHERS ARE NEVER SELECTED";
	      force_newline ();
	    end;
	    raise Exit; (* We do not need to check the others *)
	  end
	end else begin
	  (* when it is ambiguous *)
	  if !mode_choose_first_match then begin
	    if debug then begin
	      print_string "AMBIGUOUSITY CANNOT RESOLVED AUTOMATICALLY";
	      force_newline ();
	    end;
	  end
	end
      with
      | e when e <> Exit -> ()
      end;
      if debug then begin print_string "||"; force_newline () end ) konsts
  with
    Exit -> ()
  end;

  if debug then begin
    print_string "END OF TYPECASE ANALISYS";
    force_newline ();
  end;

  let anti, flow =
    match !found with
    | [] -> 
	if debug then begin
	  print_string "no match";
	  force_newline ();
	end;
	raise No_match
    | [pos,t,flows] -> 
	if debug then begin
	  print_string "1 match: ";
	  Printtyp.type_expr t;
	  force_newline ();
	end;
	let pos' = 
	  try
	    List.assoc path layout + pos
	  with
	  | Not_found -> pos
	in
	t,
	begin match flows with
	| None -> None
	| Some flows -> Some (Flow_node { flow_pos= pos'; 
					  flow_sub= flows; 
					  flow_id= -1})
	end
    | _ -> 
	let atype =
	  Dtype.anti_unify_types env 
	    (List.map (fun (_,t,_) -> t) !found)
	in

	if debug then begin
	  printf "%d matches: " (List.length !found);
	  Printtyp.type_expr atype;
	  force_newline ();
	end;

	atype, None
  in

  if debug then begin
    print_string "unify: ";
    Printtyp.type_expr typ;
    print_string " ~= ";
    Printtyp.type_expr anti;
  end;

  (* typ is copied: unification is not destructive *)
  Ctype.unify env typ anti;

  if debug then begin
    force_newline ();
    print_string " ==> ";
    Printtyp.type_expr typ;
    force_newline ();
  end;

  typ, flow

(* returns the inferred type and flow *)
and reconst_konst log env recvdescs layout path typ vdesc = 
  (* recvdescs is already updated *)
  match vdesc.val_konstraints with
  | Konst_None -> failwith("reconst_konst ???")
  | Konst_Cases (p,cks) -> 
      reconst_konst_cases log env recvdescs layout path typ cks
  | Konst_Derived (p,dks) -> (* should use different function ? *) 
      let typ, flow = 
	reconst_konst_cases log env recvdescs layout path typ 
	  [{konst_case= vdesc.val_type; konst_derived= dks}]
      in
      typ,
      match flow with
      |	Some (Flow_node ({flow_pos= 0} as f)) ->
	  Some (Flow_node {f with flow_pos = -1})
      |	None -> None
      |	_ -> assert false
	    

(* Check p : t <: \sigma K *)
(* returns the inferred type, flag for type changed or not, and flow *)
and reconst_path log env recvdescs layout typ path vdesc =

  if debug then begin
    print_string "?? ";
    Printtyp.path path;
    print_string " : ";
    Printtyp.type_expr typ;
(*
    print_string " <: ";
    Printtyp.konstraints vdesc.val_konstraints;
*)
  
    new_par ();
  end;

  if vdesc.val_konstraints = Konst_None then begin
    if debug then begin
      end_par ();
      print_string "This is generic primitive";
      force_newline ();
    end;
    (* we need run time types *)
    let copy_cache = create_copy_cache Duplicate in
    let typ = copy_type copy_cache typ in
    let vtype = copy_type copy_cache vdesc.val_type in
    let dvars = generalized_dynamic_variables vtype in
    begin try
      if debug then begin
	Printtyp.type_scheme vtype;
	Format.print_string " ~=~ ";
	Printtyp.type_scheme typ;
	Format.print_newline ();
      end;
      (* typ and vtype are copied. This unification is not destructive *)
      Ctype.unify env typ vtype
    with
      _ -> raise (Generic_system_error "Unification must suceed!")
    end;
    typ, false, Some (Flow_type dvars)

  end else begin
    (* non primitives *)

    let copy_cache = create_copy_cache Duplicate in
    let typ = copy_type copy_cache typ in

    let my_id = !node_id in
    incr node_id;
    try
      let t,flow = 
	let rec assoc_cache (path,typ) = function
	  | ((p,t),id) :: xs ->
	      if path = p then begin
(*
  print_string "find ";
  Printtyp.path path;
  print_string ",";
  Printtyp.type_expr typ;
  print_string " = ";
  Printtyp.path p;
  print_string ",";
  Printtyp.type_expr t;
  force_newline ();
*)
		if Ctype.equal env false [typ] [t] then id
		else assoc_cache (path,typ) xs
	      end else assoc_cache (path,typ) xs
	  | [] -> raise Not_found
	in
	try
	  let id = assoc_cache (path, typ) log.cache in
	  if debug then begin
	    print_string "FOUND IN THE CACHE";
	    force_newline ();
	  end;
	  typ, Some (Flow_loop id)
	with
	| Not_found ->
	    let log = 
	      let paths, rec_check =
	      	if List.mem path log.paths then log.paths, true
  	      	else path :: log.paths, false
	      in
	      { paths= paths;
	      	type_grown= log.type_grown;
	      	cache= ((path,typ),my_id)::log.cache }
	    in
	    reconst_konst log env ((path,vdesc)::recvdescs) 
	      layout path typ vdesc
      in
  
      begin match flow with
      |	Some (Flow_node node) -> node.flow_id <- my_id
      |	Some (Flow_loop _) -> ()
      |	None -> ()
      |	_ -> assert false
      end;

      if debug then end_par ();

      (* check that typ will be changed... *)
      (* possible bug: t may contain free variables occur in the program *)

      let type_changed = not (generic_moregeneral env t typ) in

      (* This does not change the real types *)
      Ctype.unify env t typ;

      if debug then begin
	printf "%s: " (if flow = None then "?" else "O");
	Printtyp.path path;
	print_string " : ";
	Printtyp.type_expr typ;
(*
        print_string " <: ";
        Printtyp.extended_scheme vdesc.val_type vdesc.val_konstraints;
*)
	force_newline ();
      end;

      typ, type_changed, flow
    with
    | No_match -> 
  	if debug then end_par ();
  	raise (No_match_with (path, vdesc, typ))
  end
;;

(* returns the inferred type, flag for type changed or not, and flow *)
let reconst_exp e =
  let path, vdesc = get_ident e in
  let layout = Ktype.generic_compilation_layout vdesc.val_konstraints in
  try
    reconst_path empty_log e.exp_env [] layout e.exp_type path vdesc 
  with
    No_match_with (p,v,t) ->
      raise (Invalid_full_instance (e,p,v,t))
  
let reconst exps = 
  let changed, flows_rev, unifpairs = 
    List.fold_left (fun (ch,frev,unifpairs) e -> 
      let t',changed, flow = reconst_exp e in
      ch || changed, flow :: frev, (e, t') :: unifpairs) 
      (false,[],[]) exps
  in
  List.iter (fun (e,t') -> 
    try
      Ctype.unify e.exp_env e.exp_type t'
    with
    | Ctype.Unify trace -> raise (Generic_expr_type_clash (e.exp_loc,trace)))
    unifpairs;
  changed, List.rev flows_rev 
;;

let reconst_exps exps =
  node_id := 0;
  mode_choose_first_match := true;
  let prev_amb = ref None in

  if debug then new_par ();
  let changed, flows = reconst exps in 

  if debug then begin
    end_par ();
    print_newline ();
  end;

  if debug2 then begin
    if exps <> [] then begin
      List.iter (fun exp ->
	let path, vdesc = get_ident exp in
	Printtyp.path path;
	print_string " : ";
	Printtyp.type_expr exp.exp_type;
	print_newline () ) exps;
      print_string "---";
      print_newline ();
    end
  end;

  flows
;;

let unify_free_dvar_to_dyn exps =
  let modified = ref false in
  List.iter (fun e ->
    let path, vdesc = get_ident e in

    let rec unifier ty =
      begin match ty.desc with
      | Tdvar (x,y) when !x = Not_generalized && !y = Not_exportable ->
	  (* This changes the real types *)
	  Ctype.unify e.exp_env ty Predef.type_dyn;
	  modified := true
      | _ -> ()
      end;
      Btype.iter_type_expr unifier ty
    in
    unifier e.exp_type) exps;
  !modified
;;

let reconst_str env strs =
  let gabstract = find_generics_str env strs in

  if gabstract <> [] then begin
    if debug || debug2 then begin
      Format.print_string "GENERIC TYPING";
      Format.print_newline ();
  
      Format.print_string "ASSIGN GABSTRACT----------";
      Format.print_newline ();
    end;

    let fullinsts = assign gabstract in
  
    if fullinsts <> [] then begin
      let flows =

  	if debug || debug2 then begin
	  Format.print_string "FIRST PHASE----------";
  	  Format.print_newline ();
	end;
	let flows = reconst_exps fullinsts in

  	ignore begin
	  try
	    unify_free_dvar_to_dyn fullinsts
	  with
	  | e -> 
	      print_string "unify_free_dvar_to_dyn: ";
	      print_string (Printexc.to_string e);
	      force_newline ();
	      raise e
	end;
  	if debug || debug2 then begin
	  Format.print_string "SECOND LOOP (Final check with DYN unification)----------";
  	  Format.print_newline ();
	end;
  	reconst_exps fullinsts;
      in
    
      List.iter2 (fun exp flow ->
  	match exp.exp_desc with
  	| Texp_ident (_,_,flowref) -> 
  	    begin match flow with
  	    | Some flow -> flowref := Inst_full_generic flow
  	    | None -> assert false
  	    end
  	| _ -> assert false)  fullinsts flows;
    
    
      if debug then begin
	List.iter2 (fun exp flow ->
  	  let p, vdesc = get_ident exp in
  	  let flow =
  	    match flow with
  	    | Some flow -> flow
  	    | None -> assert false
  	  in
  	  Printtyp.path p;
  	  Format.print_string " : ";
  	  Printtyp.type_expr exp.exp_type;
	  Format.print_newline ();
  	  Format.print_string "    ";
  	  Format.open_box 0;
  	  print_flow flow;
  	  Format.close_box ();
  	  Format.force_newline ()) fullinsts flows;
    
	Format.print_string "Generic typing finished";
	Format.print_newline ();
      end;

    end
  end
;;

let reconst_str env strs =
  Ctype.under_generic_typing := true;
  try
    let result = reconst_str env strs in
    Ctype.under_generic_typing := false;
    result
  with
    e ->
      Ctype.under_generic_typing := false;
      raise e

